from __future__ import annotations

from math import sqrt

import pytest

try:
    from tqdm.autonotebook import tqdm

    from monty.multiprocessing import imap_tqdm
except ImportError:
    tqdm = None


@pytest.mark.skipif(tqdm is None, reason="tqdm not installed")
def test_imap_tqdm():
    results = imap_tqdm(4, sqrt, range(10_000))
    assert len(results) == 10_000
    assert results[0] == 0
    assert results[400] == 20
    assert results[9999] == 99.99499987499375
    results = imap_tqdm(4, sqrt, (i**2 for i in range(10_000)))
    assert len(results) == 10_000
    assert results[0] == 0
    assert results[400] == 400



from __future__ import annotations

import pytest

from monty.fractions import gcd, gcd_float, lcm


def test_gcd():
    assert gcd(7, 14, 63) == 7


def test_lcm():
    assert lcm(2, 3, 4) == 12


def test_gcd_float():
    vs = [6.2, 12.4, 15.5 + 5e-9]
    assert gcd_float(vs, 1e-8) == pytest.approx(3.1)



from __future__ import annotations

import platform
import time

import pytest

from monty.functools import (
    TimeoutError,
    lazy_property,
    prof_main,
    return_if_raise,
    return_none_if_raise,
    timeout,
)


class TestLazy:
    def test_evaluate(self):
        # Lazy attributes should be evaluated when accessed.
        called = []

        class Foo:
            @lazy_property
            def foo(self):
                called.append("foo")
                return 1

        f = Foo()
        assert f.foo == 1
        assert len(called) == 1

    def test_evaluate_once(self):
        # lazy_property attributes should be evaluated only once.
        called = []

        class Foo:
            @lazy_property
            def foo(self):
                called.append("foo")
                return 1

        f = Foo()
        assert f.foo == 1
        assert f.foo == 1
        assert f.foo == 1
        assert len(called) == 1

    def test_private_attribute(self):
        # It should be possible to create private, name-mangled
        # lazy_property attributes.
        called = []

        class Foo:
            @lazy_property
            def __foo(self):
                called.append("foo")
                return 1

            def get_foo(self):
                return self.__foo

        f = Foo()
        assert f.get_foo() == 1
        assert f.get_foo() == 1
        assert f.get_foo() == 1
        assert len(called) == 1

    def test_reserved_attribute(self):
        # It should be possible to create reserved lazy_property attributes.
        called = []

        class Foo:
            @lazy_property
            def __foo__(self):
                called.append("foo")
                return 1

        f = Foo()
        assert f.__foo__ == 1
        assert f.__foo__ == 1
        assert f.__foo__ == 1
        assert len(called) == 1

    def test_result_shadows_descriptor(self):
        # The result of the function call should be stored in
        # the object __dict__, shadowing the descriptor.
        called = []

        class Foo:
            @lazy_property
            def foo(self):
                called.append("foo")
                return 1

        f = Foo()
        assert isinstance(Foo.foo, lazy_property)
        assert f.foo is f.foo
        assert f.foo is f.__dict__["foo"]  # !
        assert len(called) == 1

        assert f.foo == 1
        assert f.foo == 1
        assert len(called) == 1

        lazy_property.invalidate(f, "foo")

        assert f.foo == 1
        assert len(called) == 2

        assert f.foo == 1
        assert f.foo == 1
        assert len(called) == 2

    def test_readonly_object(self):
        # The descriptor should raise an AttributeError when lazy_property is
        # used on a read-only object (an object with __slots__).
        called = []

        class Foo:
            __slots__ = ()

            @lazy_property
            def foo(self):
                called.append("foo")
                return 1

        f = Foo()
        assert len(called) == 0

        with pytest.raises(
            AttributeError, match="'Foo' object has no attribute '__dict__'"
        ):
            f.foo

        # The function was not called
        assert len(called) == 0

    def test_introspection(self):
        # The lazy_property decorator should support basic introspection.

        class Foo:
            def foo(self):
                """foo func doc"""

            @lazy_property
            def bar(self):
                """bar func doc"""

        assert Foo.foo.__name__ == "foo"
        assert Foo.foo.__doc__ == "foo func doc"
        assert "test_functools" in Foo.foo.__module__

        assert Foo.bar.__name__ == "bar"
        assert Foo.bar.__doc__ == "bar func doc"
        assert "test_functools" in Foo.bar.__module__


class TestInvalidate:
    def test_invalidate_attribute(self):
        # It should be possible to invalidate a lazy_property attribute.
        called = []

        class Foo:
            @lazy_property
            def foo(self):
                called.append("foo")
                return 1

        f = Foo()
        assert f.foo == 1
        assert len(called) == 1

        lazy_property.invalidate(f, "foo")

        assert f.foo == 1
        assert len(called) == 2

    def test_invalidate_attribute_twice(self):
        # It should be possible to invalidate a lazy_property attribute
        # twice without causing harm.
        called = []

        class Foo:
            @lazy_property
            def foo(self):
                called.append("foo")
                return 1

        f = Foo()
        assert f.foo == 1
        assert len(called) == 1

        lazy_property.invalidate(f, "foo")
        lazy_property.invalidate(f, "foo")  # Nothing happens

        assert f.foo == 1
        assert len(called) == 2

    def test_invalidate_uncalled_attribute(self):
        # It should be possible to invalidate an empty attribute
        # cache without causing harm.
        called = []

        class Foo:
            @lazy_property
            def foo(self):
                called.append("foo")
                return 1

        f = Foo()
        assert len(called) == 0
        lazy_property.invalidate(f, "foo")  # Nothing happens

    def test_invalidate_private_attribute(self):
        # It should be possible to invalidate a private lazy_property attribute.
        called = []

        class Foo:
            @lazy_property
            def __foo(self):
                called.append("foo")
                return 1

            def get_foo(self):
                return self.__foo

        f = Foo()
        assert f.get_foo() == 1
        assert len(called) == 1

        lazy_property.invalidate(f, "__foo")

        assert f.get_foo() == 1
        assert len(called) == 2

    def test_invalidate_mangled_attribute(self):
        # It should be possible to invalidate a private lazy_property attribute
        # by its mangled name.
        called = []

        class Foo:
            @lazy_property
            def __foo(self):
                called.append("foo")
                return 1

            def get_foo(self):
                return self.__foo

        f = Foo()
        assert f.get_foo() == 1
        assert len(called) == 1

        lazy_property.invalidate(f, "_Foo__foo")

        assert f.get_foo() == 1
        assert len(called) == 2

    def test_invalidate_reserved_attribute(self):
        # It should be possible to invalidate a reserved lazy_property attribute.
        called = []

        class Foo:
            @lazy_property
            def __foo__(self):
                called.append("foo")
                return 1

        f = Foo()
        assert f.__foo__ == 1
        assert len(called) == 1

        lazy_property.invalidate(f, "__foo__")

        assert f.__foo__ == 1
        assert len(called) == 2

    def test_invalidate_nonlazy_attribute(self):
        # Invalidating an attribute that is not lazy_property should
        # raise an AttributeError.
        called = []

        class Foo:
            def foo(self):
                called.append("foo")
                return 1

        f = Foo()
        with pytest.raises(
            AttributeError, match="'Foo.foo' is not a lazy_property attribute"
        ):
            lazy_property.invalidate(f, "foo")

    def test_invalidate_nonlazy_private_attribute(self):
        # Invalidating a private attribute that is not lazy_property should
        # raise an AttributeError.
        called = []

        class Foo:
            def __foo(self):
                called.append("foo")
                return 1

        f = Foo()
        with pytest.raises(
            AttributeError, match="type object 'Foo' has no attribute 'foo'"
        ):
            lazy_property.invalidate(f, "foo")

    def test_invalidate_unknown_attribute(self):
        # Invalidating an unknown attribute should
        # raise an AttributeError.
        called = []

        class Foo:
            @lazy_property
            def foo(self):
                called.append("foo")
                return 1

        f = Foo()
        with pytest.raises(
            AttributeError, match="type object 'Foo' has no attribute 'bar'"
        ):
            lazy_property.invalidate(f, "bar")

    def test_invalidate_readonly_object(self):
        # Calling invalidate on a read-only object should
        # raise an AttributeError.
        called = []

        class Foo:
            __slots__ = ()

            @lazy_property
            def foo(self):
                called.append("foo")
                return 1

        f = Foo()
        with pytest.raises(
            AttributeError, match="'Foo' object has no attribute '__dict__'"
        ):
            lazy_property.invalidate(f, "foo")


# A lazy_property subclass
class cached(lazy_property):
    pass


class TestInvalidateSubclass:
    def test_invalidate_attribute(self):
        # It should be possible to invalidate a cached attribute.
        called = []

        class Bar:
            @cached
            def bar(self):
                called.append("bar")
                return 1

        b = Bar()
        assert b.bar == 1
        assert len(called) == 1

        cached.invalidate(b, "bar")

        assert b.bar == 1
        assert len(called) == 2

    def test_invalidate_attribute_twice(self):
        # It should be possible to invalidate a cached attribute
        # twice without causing harm.
        called = []

        class Bar:
            @cached
            def bar(self):
                called.append("bar")
                return 1

        b = Bar()
        assert b.bar == 1
        assert len(called) == 1

        cached.invalidate(b, "bar")
        cached.invalidate(b, "bar")  # Nothing happens

        assert b.bar == 1
        assert len(called) == 2

    def test_invalidate_uncalled_attribute(self):
        # It should be possible to invalidate an empty attribute
        # cache without causing harm.
        called = []

        class Bar:
            @cached
            def bar(self):
                called.append("bar")
                return 1

        b = Bar()
        assert len(called) == 0
        cached.invalidate(b, "bar")  # Nothing happens

    def test_invalidate_private_attribute(self):
        # It should be possible to invalidate a private cached attribute.
        called = []

        class Bar:
            @cached
            def __bar(self):
                called.append("bar")
                return 1

            def get_bar(self):
                return self.__bar

        b = Bar()
        assert b.get_bar() == 1
        assert len(called) == 1

        cached.invalidate(b, "__bar")

        assert b.get_bar() == 1
        assert len(called) == 2

    def test_invalidate_mangled_attribute(self):
        # It should be possible to invalidate a private cached attribute
        # by its mangled name.
        called = []

        class Bar:
            @cached
            def __bar(self):
                called.append("bar")
                return 1

            def get_bar(self):
                return self.__bar

        b = Bar()
        assert b.get_bar() == 1
        assert len(called) == 1

        cached.invalidate(b, "_Bar__bar")

        assert b.get_bar() == 1
        assert len(called) == 2

    def test_invalidate_reserved_attribute(self):
        # It should be possible to invalidate a reserved cached attribute.
        called = []

        class Bar:
            @cached
            def __bar__(self):
                called.append("bar")
                return 1

        b = Bar()
        assert b.__bar__ == 1
        assert len(called) == 1

        cached.invalidate(b, "__bar__")

        assert b.__bar__ == 1
        assert len(called) == 2

    def test_invalidate_uncached_attribute(self):
        # Invalidating an attribute that is not cached should
        # raise an AttributeError.
        called = []

        class Bar:
            def bar(self):
                called.append("bar")
                return 1

        b = Bar()
        with pytest.raises(AttributeError, match="'Bar.bar' is not a cached attribute"):
            cached.invalidate(b, "bar")

    def test_invalidate_uncached_private_attribute(self):
        # Invalidating a private attribute that is not cached should
        # raise an AttributeError.
        called = []

        class Bar:
            def __bar(self):
                called.append("bar")
                return 1

        b = Bar()
        with pytest.raises(
            AttributeError, match="'Bar._Bar__bar' is not a cached attribute"
        ):
            cached.invalidate(
                b,
                "__bar",
            )

    def test_invalidate_unknown_attribute(self):
        # Invalidating an unknown attribute should
        # raise an AttributeError.
        called = []

        class Bar:
            @cached
            def bar(self):
                called.append("bar")
                return 1

        b = Bar()
        with pytest.raises(
            AttributeError, match="type object 'Bar' has no attribute 'baz'"
        ):
            lazy_property.invalidate(
                b,
                "baz",
            )

    def test_invalidate_readonly_object(self):
        # Calling invalidate on a read-only object should
        # raise an AttributeError.
        called = []

        class Bar:
            __slots__ = ()

            @cached
            def bar(self):
                called.append("bar")
                return 1

        b = Bar()
        with pytest.raises(
            AttributeError, match="'Bar' object has no attribute '__dict__'"
        ):
            cached.invalidate(
                b,
                "bar",
            )

    def test_invalidate_superclass_attribute(self):
        # cached.invalidate CANNOT invalidate a superclass (lazy_property) attribute.
        called = []

        class Bar:
            @lazy_property
            def bar(self):
                called.append("bar")
                return 1

        b = Bar()
        with pytest.raises(AttributeError, match="'Bar.bar' is not a cached attribute"):
            cached.invalidate(
                b,
                "bar",
            )

    def test_invalidate_subclass_attribute(self):
        # Whereas lazy_property.invalidate CAN invalidate a subclass (cached) attribute.
        called = []

        class Bar:
            @cached
            def bar(self):
                called.append("bar")
                return 1

        b = Bar()
        assert b.bar == 1
        assert len(called) == 1

        lazy_property.invalidate(b, "bar")

        assert b.bar == 1
        assert len(called) == 2


class TestTryOrReturn:
    def test_decorator(self):
        class A:
            @return_if_raise(ValueError, "hello")
            def return_one(self):
                return 1

            @return_if_raise(ValueError, "hello")
            def return_hello(self):
                raise ValueError

            @return_if_raise(KeyError, "hello")
            def reraise_value_error(self):
                raise ValueError

            @return_if_raise([KeyError, ValueError], "hello")
            def catch_exc_list(self):
                import random

                if random.randint(0, 1) == 0:
                    raise ValueError
                raise KeyError

            @return_none_if_raise(TypeError)
            def return_none(self):
                raise TypeError

        a = A()
        assert a.return_one() == 1
        assert a.return_hello() == "hello"
        with pytest.raises(ValueError):
            a.reraise_value_error()
        assert a.catch_exc_list() == "hello"
        assert a.return_none() is None


class TestTimeout:
    @pytest.mark.skipif(platform.system() == "Windows", reason="Skip on windows")
    def test_with(self):
        try:
            with timeout(1, "timeout!"):
                time.sleep(2)
            raise AssertionError("Did not timeout!")
        except TimeoutError as ex:
            assert ex.message == "timeout!"


class TestProfMain:
    def test_prof_decorator(self):
        """Testing prof_main decorator."""
        import sys

        @prof_main
        def main():
            return sys.exit(1)

        # Have to change argv before calling main.
        # Will restore original values before returning.
        _ = sys.argv[:]
        if len(sys.argv) == 1:
            sys.argv.append("prof")
        else:
            sys.argv[1] = "prof"



from __future__ import annotations

import random

from monty.string import remove_non_ascii


def test_remove_non_ascii():
    s = "".join(chr(random.randint(0, 127)) for _ in range(10))
    s += "".join(chr(random.randint(128, 150)) for _ in range(10))
    clean = remove_non_ascii(s)
    assert len(clean) == 10



from __future__ import annotations

import glob
import json
import os

import pytest

from monty.serialization import dumpfn, loadfn
from monty.tempfile import ScratchDir

try:
    import msgpack
except ImportError:
    msgpack = None


class TestSerial:
    @classmethod
    def teardown_class(cls):
        # Cleans up test files if a test fails
        files_to_clean_up = glob.glob("monte_test.*")
        for fn in files_to_clean_up:
            os.remove(fn)

    def test_dumpfn_loadfn(self):
        d = {"hello": "world"}

        # Test standard configuration
        for ext in (
            "json",
            "yaml",
            "yml",
            "json.gz",
            "yaml.gz",
            "json.bz2",
            "yaml.bz2",
        ):
            fn = f"monte_test.{ext}"
            dumpfn(d, fn)
            d2 = loadfn(fn)
            assert d == d2, f"Test file with extension {ext} did not parse correctly"
            os.remove(fn)

        # Test custom kwarg configuration
        dumpfn(d, "monte_test.json", indent=4)
        d2 = loadfn("monte_test.json")
        assert d == d2
        os.remove("monte_test.json")
        dumpfn(d, "monte_test.yaml")
        d2 = loadfn("monte_test.yaml")
        assert d == d2
        os.remove("monte_test.yaml")

        # Check if fmt override works.
        dumpfn(d, "monte_test.json", fmt="yaml")
        with pytest.raises(json.decoder.JSONDecodeError):
            loadfn("monte_test.json")
        d2 = loadfn("monte_test.json", fmt="yaml")
        assert d == d2
        os.remove("monte_test.json")

        with pytest.raises(TypeError):
            dumpfn(d, "monte_test.txt", fmt="garbage")
        with pytest.raises(TypeError):
            loadfn("monte_test.txt", fmt="garbage")

    @pytest.mark.skipif(msgpack is None, reason="msgpack-python not installed.")
    def test_mpk(self):
        d = {"hello": "world"}

        # Test automatic format detection
        dumpfn(d, "monte_test.mpk")
        d2 = loadfn("monte_test.mpk")
        assert d == d2
        os.remove("monte_test.mpk")

        # Test to ensure basename is respected, and not directory
        with ScratchDir("."):
            os.mkdir("mpk_test")
            os.chdir("mpk_test")
            fname = os.path.abspath("test_file.json")
            dumpfn({"test": 1}, fname)
            with open("test_file.json", encoding="utf-8") as f:
                reloaded = json.loads(f.read())
            assert reloaded["test"] == 1



from __future__ import annotations

from collections import UserDict

import pytest

from monty.collections import (
    AttrDict,
    ControlledDict,
    FrozenAttrDict,
    MongoDict,
    Namespace,
    dict2namedtuple,
    frozendict,
    tree,
)


def test_tree():
    x = tree()
    x["a"]["b"]["c"]["d"] = 1
    assert "b" in x["a"]
    assert "c" not in x["a"]
    assert "c" in x["a"]["b"]
    assert x["a"]["b"]["c"]["d"] == 1


class TestControlledDict:
    def test_add_allowed(self):
        dct = ControlledDict(a=1)
        dct._allow_add = True

        dct["b"] = 2
        assert dct["b"] == 2

        dct.update(d=3)
        assert dct["d"] == 3

        dct.setdefault("e", 4)
        assert dct["e"] == 4

    def test_add_disabled(self):
        dct = ControlledDict(a=1)
        dct._allow_add = False

        with pytest.raises(TypeError, match="add is disabled"):
            dct["b"] = 2

        with pytest.raises(TypeError, match="add is disabled"):
            dct.update(b=2)

        with pytest.raises(TypeError, match="add is disabled"):
            dct.setdefault("c", 2)

    def test_update_allowed(self):
        dct = ControlledDict(a=1)
        dct._allow_update = True

        dct["a"] = 2
        assert dct["a"] == 2

        dct.update({"a": 3})
        assert dct["a"] == 3

        # Test Iterator handling
        dct.update(zip(["c", "d"], [11, 12]))
        assert dct["c"] == 11

        dct.setdefault("a", 4)  # existing key
        assert dct["a"] == 3

    def test_update_disabled(self):
        dct = ControlledDict(a=1)
        dct._allow_update = False

        with pytest.raises(TypeError, match="update is disabled"):
            dct["a"] = 2

        with pytest.raises(TypeError, match="update is disabled"):
            dct.update({"a": 3})

        with pytest.raises(TypeError, match="update is disabled"):
            dct.setdefault("a", 4)

    def test_del_allowed(self):
        dct = ControlledDict(a=1, b=2, c=3, d=4)
        dct._allow_del = True

        del dct["a"]
        assert "a" not in dct

        val = dct.pop("b")
        assert val == 2 and "b" not in dct

        val = dct.popitem()
        assert val == ("c", 3) and "c" not in dct

        dct.clear()
        assert dct == {}

    def test_del_disabled(self):
        dct = ControlledDict(a=1)
        dct._allow_del = False

        with pytest.raises(TypeError, match="delete is disabled"):
            del dct["a"]

        with pytest.raises(TypeError, match="delete is disabled"):
            dct.pop("a")

        with pytest.raises(TypeError, match="delete is disabled"):
            dct.popitem()

        with pytest.raises(TypeError, match="delete is disabled"):
            dct.clear()

    def test_frozen_like(self):
        """Make sure add and update are allowed at init time."""
        ControlledDict._allow_add = False
        ControlledDict._allow_update = False

        dct = ControlledDict({"hello": "world"})
        assert isinstance(dct, UserDict)
        assert dct["hello"] == "world"

        assert not dct._allow_add
        assert not dct._allow_update

    def test_iterator_handling(self):
        """Make sure iterators are handling correctly."""
        c_dict = ControlledDict(zip(["c", "d"], [11, 12]))
        assert c_dict["c"] == 11


def test_frozendict():
    dct = frozendict({"hello": "world"})
    assert isinstance(dct, UserDict)
    assert dct["hello"] == "world"

    assert not dct._allow_add
    assert not dct._allow_update
    assert not dct._allow_del

    # Test setter
    with pytest.raises(TypeError, match="add is disabled"):
        dct["key"] = "val"

    # Test update
    with pytest.raises(TypeError, match="add is disabled"):
        dct.update(key="val")

    # Test pop
    with pytest.raises(TypeError, match="delete is disabled"):
        dct.pop("key")

    # Test delete
    with pytest.raises(TypeError, match="delete is disabled"):
        del dct["key"]


def test_namespace_dict():
    dct = Namespace(key="val")
    assert isinstance(dct, UserDict)

    # Test setter
    dct["hello"] = "world"
    assert dct["key"] == "val"

    # Test use `update` to add new values
    dct.update({"new_key": "new_value"})
    assert dct["new_key"] == "new_value"

    # Test add (not allowed)
    with pytest.raises(TypeError, match="update is disabled"):
        dct["key"] = "val"

    with pytest.raises(TypeError, match="update is disabled"):
        dct.update({"key": "val"})

    # Test delete (not allowed)
    with pytest.raises(TypeError, match="delete is disabled"):
        del dct["key"]


def test_attr_dict():
    dct = AttrDict(foo=1, bar=2)

    # Test get attribute
    assert dct.bar == 2
    assert dct["foo"] is dct.foo

    # Test key not found error
    with pytest.raises(KeyError, match="no-such-key"):
        dct["no-such-key"]

    # Test setter
    dct.bar = "hello"
    assert dct["bar"] == "hello"

    # Test delete
    del dct.bar
    assert "bar" not in dct

    # Test builtin dict method shadowing
    with pytest.warns(UserWarning, match="shadows dict method"):
        dct["update"] = "value"


def test_frozen_attrdict():
    dct = FrozenAttrDict({"hello": "world", 1: 2})
    assert isinstance(dct, UserDict)

    # Test attribute-like operations
    with pytest.raises(TypeError, match="does not support item assignment"):
        dct.foo = "bar"

    with pytest.raises(TypeError, match="does not support item assignment"):
        dct.hello = "new"

    with pytest.raises(TypeError, match="does not support item deletion"):
        del dct.hello

    # Test get value
    assert dct["hello"] == "world"
    assert dct.hello == "world"
    assert dct["hello"] is dct.hello  # identity check

    # Test adding item
    with pytest.raises(TypeError, match="add is disabled"):
        dct["foo"] = "bar"

    # Test modifying existing item
    with pytest.raises(TypeError, match="update is disabled"):
        dct["hello"] = "new"

    # Test update
    with pytest.raises(TypeError, match="update is disabled"):
        dct.update({"hello": "world"})

    # Test pop
    with pytest.raises(TypeError, match="delete is disabled"):
        dct.pop("hello")

    with pytest.raises(TypeError, match="delete is disabled"):
        dct.popitem()

    # Test delete
    with pytest.raises(TypeError, match="delete is disabled"):
        del dct["hello"]

    with pytest.raises(TypeError, match="delete is disabled"):
        dct.clear()


def test_mongo_dict():
    m_dct = MongoDict({"a": {"b": 1}, "x": 2})
    assert m_dct.a.b == 1
    assert m_dct.x == 2
    assert "a" in m_dct
    assert "b" in m_dct.a
    assert m_dct["a"] == {"b": 1}


def test_dict2namedtuple():
    # Init from dict
    tpl = dict2namedtuple(foo=1, bar="hello")
    assert isinstance(tpl, tuple)
    assert tpl.foo == 1 and tpl.bar == "hello"

    # Init from list of tuples
    tpl = dict2namedtuple([("foo", 1), ("bar", "hello")])
    assert isinstance(tpl, tuple)
    assert tpl[0] == 1
    assert tpl[1] == "hello"
    assert tpl[0] is tpl.foo and tpl[1] is tpl.bar



from __future__ import annotations

import os
import platform
import shutil

import pytest

from monty.tempfile import ScratchDir

TEST_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "test_files")


class TestScratchDir:
    def setup_method(self):
        self.cwd = os.getcwd()
        os.chdir(TEST_DIR)
        self.scratch_root = os.path.join(TEST_DIR, "..", "..", "tempscratch")
        os.mkdir(self.scratch_root)

    def test_with_copy(self):
        # We write a pre-scratch file.
        with open("pre_scratch_text", "w", encoding="utf-8") as f:
            f.write("write")

        with ScratchDir(
            self.scratch_root,
            copy_from_current_on_enter=True,
            copy_to_current_on_exit=True,
        ) as d:
            with open("scratch_text", "w", encoding="utf-8") as f:
                f.write("write")
            files = os.listdir(d)
            assert "scratch_text" in files
            assert "empty_file.txt" in files
            assert "pre_scratch_text" in files

            # We remove the pre-scratch file.
            os.remove("pre_scratch_text")

        # Make sure the tempdir is deleted.
        assert not os.path.exists(d)
        files = os.listdir(".")
        assert "scratch_text" in files

        # We check that the pre-scratch file no longer exists (because it is
        # deleted in the scratch)
        assert "pre_scratch_text" not in files
        os.remove("scratch_text")

    def test_with_copy_gzip(self):
        # We write a pre-scratch file.
        with open("pre_scratch_text", "w", encoding="utf-8") as f:
            f.write("write")
        init_gz_files = [f for f in os.listdir(os.getcwd()) if f.endswith(".gz")]
        with pytest.warns(match="Both 3000_lines.txt and 3000_lines.txt.gz exist."):
            with (
                ScratchDir(
                    self.scratch_root,
                    copy_from_current_on_enter=True,
                    copy_to_current_on_exit=True,
                    gzip_on_exit=True,
                ),
                open("scratch_text", "w", encoding="utf-8") as f,
            ):
                f.write("write")
        files = os.listdir(os.getcwd())

        # Make sure the scratch_text.gz exists
        assert "scratch_text.gz" in files
        for f in files:
            if f.endswith(".gz") and f not in init_gz_files:
                os.remove(f)
        os.remove("pre_scratch_text")

    def test_with_copy_nodelete(self):
        # We write a pre-scratch file.
        with open("pre_scratch_text", "w", encoding="utf-8") as f:
            f.write("write")

        with ScratchDir(
            self.scratch_root,
            copy_from_current_on_enter=True,
            copy_to_current_on_exit=True,
            delete_removed_files=False,
        ) as d:
            with open("scratch_text", "w", encoding="utf-8") as f:
                f.write("write")
            files = os.listdir(d)
            assert "scratch_text" in files
            assert "empty_file.txt" in files
            assert "pre_scratch_text" in files

            # We remove the pre-scratch file.
            os.remove("pre_scratch_text")

        # Make sure the tempdir is deleted.
        assert not os.path.exists(d)
        files = os.listdir(".")
        assert "scratch_text" in files

        # We check that the pre-scratch file DOES still exists
        assert "pre_scratch_text" in files
        os.remove("scratch_text")
        os.remove("pre_scratch_text")

    def test_no_copy(self):
        with ScratchDir(
            self.scratch_root,
            copy_from_current_on_enter=False,
            copy_to_current_on_exit=False,
        ) as d:
            with open("scratch_text", "w", encoding="utf-8") as f:
                f.write("write")
            files = os.listdir(d)
            assert "scratch_text" in files
            assert "empty_file.txt" not in files

        # Make sure the tempdir is deleted.
        assert not os.path.exists(d)
        files = os.listdir(".")
        assert "scratch_text" not in files

    def test_symlink(self):
        if platform.system() != "Windows":
            with ScratchDir(
                self.scratch_root,
                copy_from_current_on_enter=False,
                copy_to_current_on_exit=False,
                create_symbolic_link=True,
            ) as d:
                with open("scratch_text", "w", encoding="utf-8") as f:
                    f.write("write")
                files = os.listdir(d)
                assert "scratch_text" in files
                assert "empty_file.txt" not in files

            # Make sure the tempdir is deleted.
            assert not os.path.exists(d)
            files = os.listdir(".")
            assert "scratch_text" not in files

            # Make sure the symlink is removed
            assert not os.path.islink("scratch_link")

    def test_bad_root(self):
        with ScratchDir("bad_groot") as d:
            assert d == TEST_DIR

    def teardown_method(self):
        os.chdir(self.cwd)
        shutil.rmtree(self.scratch_root)



from __future__ import annotations

from monty.subprocess import Command


def test_command():
    """Test Command class"""
    sleep05 = Command("sleep 0.5")

    sleep05.run(timeout=1)
    # DEBUG: this unit test fail in Windows CI intermittently (PR702)
    full_msg = f"{sleep05=}\n{sleep05.error=}\n{sleep05.output}"
    assert sleep05.retcode == 0, full_msg
    assert not sleep05.killed

    sleep05.run(timeout=0.1)
    assert sleep05.retcode != 0
    assert sleep05.killed



from __future__ import annotations

import os
import platform
import shutil
import tempfile
from gzip import GzipFile
from pathlib import Path

import pytest

from monty.shutil import (
    compress_dir,
    compress_file,
    copy_r,
    decompress_dir,
    decompress_file,
    gzip_dir,
    remove,
)

TEST_DIR = os.path.join(os.path.dirname(__file__), "test_files")


class TestCopyR:
    def setup_method(self):
        os.mkdir(os.path.join(TEST_DIR, "cpr_src"))
        with open(
            os.path.join(TEST_DIR, "cpr_src", "test"), "w", encoding="utf-8"
        ) as f:
            f.write("what")
        os.mkdir(os.path.join(TEST_DIR, "cpr_src", "sub"))
        with open(
            os.path.join(TEST_DIR, "cpr_src", "sub", "testr"), "w", encoding="utf-8"
        ) as f:
            f.write("what2")
        if platform.system() != "Windows":
            os.symlink(
                os.path.join(TEST_DIR, "cpr_src", "test"),
                os.path.join(TEST_DIR, "cpr_src", "mysymlink"),
            )

    def test_recursive_copy_and_compress(self):
        copy_r(os.path.join(TEST_DIR, "cpr_src"), os.path.join(TEST_DIR, "cpr_dst"))
        assert os.path.exists(os.path.join(TEST_DIR, "cpr_dst", "test"))
        assert os.path.exists(os.path.join(TEST_DIR, "cpr_dst", "sub", "testr"))

        compress_dir(os.path.join(TEST_DIR, "cpr_src"))
        assert os.path.exists(os.path.join(TEST_DIR, "cpr_src", "test.gz"))
        assert os.path.exists(os.path.join(TEST_DIR, "cpr_src", "sub", "testr.gz"))

        decompress_dir(os.path.join(TEST_DIR, "cpr_src"))
        assert os.path.exists(os.path.join(TEST_DIR, "cpr_src", "test"))
        assert os.path.exists(os.path.join(TEST_DIR, "cpr_src", "sub", "testr"))
        with open(os.path.join(TEST_DIR, "cpr_src", "test"), encoding="utf-8") as f:
            txt = f.read()
            assert txt == "what"

    def test_pathlib(self):
        test_path = Path(TEST_DIR)
        copy_r(test_path / "cpr_src", test_path / "cpr_dst")
        assert (test_path / "cpr_dst" / "test").exists()
        assert (test_path / "cpr_dst" / "sub" / "testr").exists()

    def teardown_method(self):
        shutil.rmtree(os.path.join(TEST_DIR, "cpr_src"))
        shutil.rmtree(os.path.join(TEST_DIR, "cpr_dst"))


class TestCompressFileDir:
    def setup_method(self):
        with open(os.path.join(TEST_DIR, "tempfile"), "w", encoding="utf-8") as f:
            f.write("hello world")

    def test_compress_and_decompress_file(self):
        fname = os.path.join(TEST_DIR, "tempfile")

        for fmt in ["gz", "bz2"]:
            compress_file(fname, fmt)
            assert os.path.exists(fname + "." + fmt)
            assert not os.path.exists(fname)

            decompress_file(fname + "." + fmt)
            assert os.path.exists(fname)
            assert not os.path.exists(fname + "." + fmt)

            with open(fname, encoding="utf-8") as f:
                assert f.read() == "hello world"

        with pytest.raises(ValueError):
            compress_file("whatever", "badformat")

        # test decompress non-existent/non-compressed file
        assert decompress_file("non-existent") is None
        assert decompress_file("non-existent.gz") is None
        assert decompress_file("non-existent.bz2") is None

    def test_compress_and_decompress_with_target_dir(self):
        fname = os.path.join(TEST_DIR, "tempfile")
        target_dir = os.path.join(TEST_DIR, "temp_target_dir")

        for fmt in ["gz", "bz2"]:
            compress_file(fname, fmt, target_dir)
            compressed_file_path = os.path.join(
                target_dir, f"{os.path.basename(fname)}.{fmt}"
            )
            assert os.path.exists(compressed_file_path)
            assert not os.path.exists(fname)

            decompress_file(compressed_file_path, target_dir)
            decompressed_file_path = os.path.join(target_dir, os.path.basename(fname))
            assert os.path.exists(decompressed_file_path)
            assert not os.path.exists(compressed_file_path)

            # Reset temp file position
            shutil.move(decompressed_file_path, fname)
            shutil.rmtree(target_dir)

            with open(fname, encoding="utf-8") as f:
                assert f.read() == "hello world"

    def teardown_method(self):
        os.remove(os.path.join(TEST_DIR, "tempfile"))


class TestGzipDir:
    def setup_method(self):
        os.mkdir(os.path.join(TEST_DIR, "gzip_dir"))
        with open(
            os.path.join(TEST_DIR, "gzip_dir", "tempfile"), "w", encoding="utf-8"
        ) as f:
            f.write("what")

        self.mtime = os.path.getmtime(os.path.join(TEST_DIR, "gzip_dir", "tempfile"))

    def test_gzip_dir(self):
        full_f = os.path.join(TEST_DIR, "gzip_dir", "tempfile")
        gzip_dir(os.path.join(TEST_DIR, "gzip_dir"))

        assert os.path.exists(f"{full_f}.gz")
        assert not os.path.exists(full_f)

        with GzipFile(f"{full_f}.gz") as g:
            assert g.readline().decode("utf-8") == "what"

        assert os.path.getmtime(f"{full_f}.gz") == pytest.approx(self.mtime, 4)

    def test_gzip_dir_file_coexist(self):
        """Test case where both file and file.gz exist."""
        full_f = os.path.join(TEST_DIR, "gzip_dir", "temptestfile")
        gz_f = f"{full_f}.gz"

        # Create both the file and its gzipped version
        with open(full_f, "w", encoding="utf-8") as f:
            f.write("not gzipped")
        with GzipFile(gz_f, "wb") as g:
            g.write(b"gzipped")

        with pytest.warns(
            UserWarning, match="Both temptestfile and temptestfile.gz exist."
        ):
            gzip_dir(os.path.join(TEST_DIR, "gzip_dir"))

        # Verify contents of the files
        with open(full_f, "r", encoding="utf-8") as f:
            assert f.read() == "not gzipped"

        with GzipFile(gz_f, "rb") as g:
            assert g.read() == b"gzipped"

    def test_handle_sub_dirs(self):
        sub_dir = os.path.join(TEST_DIR, "gzip_dir", "sub_dir")
        sub_file = os.path.join(sub_dir, "new_tempfile")
        os.mkdir(sub_dir)
        with open(sub_file, "w", encoding="utf-8") as f:
            f.write("anotherwhat")

        gzip_dir(os.path.join(TEST_DIR, "gzip_dir"))

        assert os.path.exists(f"{sub_file}.gz")
        assert not os.path.exists(sub_file)

        with GzipFile(f"{sub_file}.gz") as g:
            assert g.readline().decode("utf-8") == "anotherwhat"

    def teardown_method(self):
        shutil.rmtree(os.path.join(TEST_DIR, "gzip_dir"))


class TestRemove:
    @pytest.mark.skipif(platform.system() == "Windows", reason="Skip on windows")
    def test_remove_file(self):
        tempdir = tempfile.mkdtemp(dir=TEST_DIR)
        tempf = tempfile.mkstemp(dir=tempdir)[1]
        remove(tempf)
        assert not os.path.isfile(tempf)
        shutil.rmtree(tempdir)

    @pytest.mark.skipif(platform.system() == "Windows", reason="Skip on windows")
    def test_remove_folder(self):
        tempdir = tempfile.mkdtemp(dir=TEST_DIR)
        remove(tempdir)
        assert not os.path.isdir(tempdir)

    @pytest.mark.skipif(platform.system() == "Windows", reason="Skip on windows")
    def test_remove_symlink(self):
        tempdir = tempfile.mkdtemp(dir=TEST_DIR)
        tempf = tempfile.mkstemp(dir=tempdir)[1]

        os.symlink(tempdir, os.path.join(TEST_DIR, "temp_link"))
        templink = os.path.join(TEST_DIR, "temp_link")
        remove(templink)
        assert os.path.isfile(tempf)
        assert os.path.isdir(tempdir)
        assert not os.path.islink(templink)
        remove(tempdir)

    @pytest.mark.skipif(platform.system() == "Windows", reason="Skip on windows")
    def test_remove_symlink_follow(self):
        tempdir = tempfile.mkdtemp(dir=TEST_DIR)
        tempf = tempfile.mkstemp(dir=tempdir)[1]

        os.symlink(tempdir, os.path.join(TEST_DIR, "temp_link"))
        templink = os.path.join(TEST_DIR, "temp_link")
        remove(templink, follow_symlink=True)
        assert not os.path.isfile(tempf)
        assert not os.path.isdir(tempdir)
        assert not os.path.islink(templink)



from __future__ import annotations

import bz2
import gzip
import os
import warnings
from pathlib import Path

import pytest

from monty.io import (
    EncodingWarning,
    FileLock,
    FileLockException,
    _get_line_ending,
    reverse_readfile,
    reverse_readline,
    zopen,
)
from monty.tempfile import ScratchDir

TEST_DIR = os.path.join(os.path.dirname(__file__), "test_files")


class TestGetLineEnding:
    @pytest.mark.parametrize("l_end", ["\n", "\r\n"])
    def test_get_line_ending(self, l_end):
        """Test files with:
        Unix line ending (\n).
        Windows line ending (\r\n).

        For:
            - Text file: both text mode and binary mode
            - gzip file and bzip2 file
        """
        test_file = "test_l_end.txt"
        test_line = f"This is a test{l_end}Second line{l_end}".encode()

        with ScratchDir("."):
            with open(test_file, "wb") as f:
                f.write(test_line)

            assert _get_line_ending(test_file) == l_end
            assert _get_line_ending(Path(test_file)) == l_end

            # Test text mode
            with open(test_file, "r", encoding="utf-8") as f:
                start_pos = f.tell()
                assert _get_line_ending(f) == l_end
                assert f.tell() == start_pos

            # Test binary mode
            with open(test_file, "rb") as f:
                start_pos = f.tell()
                assert _get_line_ending(f) == l_end
                assert f.tell() == start_pos

            # Test gzip file
            gzip_filename = f"{test_file}.gz"
            with gzip.open(gzip_filename, "wb") as f:
                f.write(test_line)

            # Opened file stream
            with gzip.open(gzip_filename, "rb") as f:
                start_pos = f.tell()
                assert _get_line_ending(f) == l_end
                assert f.tell() == start_pos

            # Filename directly
            assert _get_line_ending(gzip_filename) == l_end

            # Test bzip2 file stream
            bz2_filename = f"{test_file}.bz2"
            with bz2.open(bz2_filename, "wb") as f:
                f.write(test_line)

            # Opened file stream
            with bz2.open(bz2_filename, "rb") as f:
                start_pos = f.tell()
                assert _get_line_ending(f) == l_end
                assert f.tell() == start_pos

            # Filename directly
            assert _get_line_ending(bz2_filename) == l_end

    @pytest.mark.parametrize("l_end", ["\n", "\r\n"])
    def test_miss_last_l_end(self, l_end):
        """Make sure this still works if the last l_end is missing."""
        test_line = f"This is a test{l_end}Second line".encode()
        test_file = "test_l_end.txt"

        with ScratchDir("."):
            with open(test_file, "wb") as f:
                f.write(test_line)

            assert _get_line_ending(test_file) == l_end

    def test_unknown_file_type(self):
        unknown_file = 123

        with pytest.raises(TypeError, match="Unknown file type int"):
            _get_line_ending(unknown_file)

    def test_empty_file(self):
        with ScratchDir("."):
            test_file = "empty_file.txt"
            open(test_file, "w").close()

            with pytest.warns(match="File is empty, return Unix line ending \n"):
                assert _get_line_ending(test_file) == "\n"

    def test_unknown_line_ending(self):
        with ScratchDir("."):
            test_file = "test_unknown.txt"
            with open(test_file, "wb") as f:
                f.write(b"This is a test\036")

            with pytest.raises(ValueError, match="Unknown line ending"):
                _get_line_ending(test_file)


class TestReverseReadline:
    """WARNING for future coder:
    "reverse_readline" has two branches, one is the in-RAM
    reverse reading for un-supported file types or small files.
    As the default RAM threshold is "big" at around 4 MB (usually
    people just write a few lines to test), you could easily be
    testing/debugging the in-RAM branch all the time (me for example).
    """

    NUMLINES = 3000

    def test_reverse_readline(self):
        """
        We are making sure a file containing line numbers is read in reverse
        order, i.e. the first line that is read corresponds to the last line.
        number
        """
        # Test text mode
        with open(
            os.path.join(TEST_DIR, "3000_lines.txt"), encoding="utf-8", newline=""
        ) as f:
            for idx, line in enumerate(reverse_readline(f)):
                assert isinstance(line, str)
                assert line == f"{str(self.NUMLINES - idx)}{os.linesep}"

        # Test binary mode
        with open(os.path.join(TEST_DIR, "3000_lines.txt"), mode="rb") as f:
            for idx, line in enumerate(reverse_readline(f)):
                assert line == f"{str(self.NUMLINES - idx)}{os.linesep}"

    @pytest.mark.parametrize("l_end", ["\n", "\r\n"])
    def test_big_file(self, l_end):
        """
        Test read big file.

        A file of 300,000 lines is about 2 MB, but the default max_mem
        is still around 4 MB, so we have to reduce it.
        """
        file_name = "big_file.txt"
        num_lines = 300_000

        with ScratchDir("."):
            # Write test file (~ 2 MB)
            with open(file_name, "wb") as file:
                for num in range(1, num_lines + 1):
                    file.write(f"{num}{l_end}".encode())

            assert os.path.getsize(file_name) > 1_000_000  # 1 MB

            # Test text mode
            with open(file_name, mode="r", encoding="utf-8", newline="") as file:
                for idx, line in enumerate(reverse_readline(file, max_mem=4096)):
                    assert line == f"{str(num_lines - idx)}{l_end}"

            # Test binary mode
            with open(file_name, mode="rb") as file:
                for idx, line in enumerate(reverse_readline(file, max_mem=4096)):
                    assert line == f"{str(num_lines - idx)}{l_end}"

    def test_read_bz2(self):
        """
        Make sure a file containing line numbers is read in reverse order,
        i.e. the first line that is read corresponds to the last line number.
        """
        lines = []
        with zopen(os.path.join(TEST_DIR, "myfile_bz2.bz2"), "rb") as f:
            for line in reverse_readline(f):
                lines.append(line)
        assert lines == ["\n", "HelloWorld.\n"]  # test file has one empty line

    def test_read_empty_file(self):
        """
        Make sure an empty file does not throw an error when reverse_readline
        is called, which was a problem with an earlier implementation.
        """
        with pytest.warns(match="File is empty, return Unix line ending \n."):
            with open(os.path.join(TEST_DIR, "empty_file.txt"), encoding="utf-8") as f:
                for _line in reverse_readline(f):
                    pytest.fail("No error should be thrown.")

    @pytest.mark.parametrize("ram", [4, 4096, 4_0000_000])
    @pytest.mark.parametrize("l_end", ["\n", "\r\n"])
    def test_read_file_with_empty_lines(self, l_end, ram):
        """Empty lines should not be skipped.
        Using a very small RAM size to force non in-RAM mode.
        """
        contents = (f"line1{l_end}", f"{l_end}", f"line3{l_end}")
        filename = "test_empty_line.txt"

        with ScratchDir("."), warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="max_mem=4 smaller than blk_size=4096"
            )
            # Test text file
            with open(filename, "wb") as file:
                for line in contents:
                    file.write(line.encode())

            with open(filename, mode="r", newline="") as file:
                revert_contents = tuple(reverse_readline(file, max_mem=ram))
            assert revert_contents[::-1] == contents

            # Test gzip file
            gzip_filename = f"{filename}.gz"
            with gzip.open(gzip_filename, "w") as file_out:
                for line in contents:
                    file_out.write(line.encode())

            with gzip.open(gzip_filename) as g_file:
                revert_contents_gzip = tuple(reverse_readline(g_file))
            assert revert_contents_gzip[::-1] == contents

            # Test bzip2 file
            bz2_filename = f"{filename}.bz2"
            with bz2.open(bz2_filename, "w") as file_out:
                for line in contents:
                    file_out.write(line.encode())

            with bz2.open(bz2_filename) as b_file:
                revert_contents_bz2 = tuple(reverse_readline(b_file))
            assert revert_contents_bz2[::-1] == contents

    @pytest.mark.parametrize("ram", [4, 4096, 4_0000_000])
    @pytest.mark.parametrize("l_end", ["\n", "\r\n"])
    def test_different_line_endings(self, l_end, ram):
        """Using a very small RAM size to force non in-RAM mode."""
        contents = (f"Line1{l_end}", f"Line2{l_end}", f"Line3{l_end}")
        file_name = "test_file.txt"

        with ScratchDir("."), warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="max_mem=4 smaller than blk_size=4096"
            )

            with open(file_name, "wb") as file:
                for line in contents:
                    file.write(line.encode())

            # Test text mode
            with open(file_name, "r", encoding="utf-8") as file:
                for idx, line in enumerate(reverse_readline(file, max_mem=ram)):
                    # OS would automatically change line ending in text mode
                    assert (
                        line.rstrip(os.linesep) + l_end
                        == contents[len(contents) - idx - 1]
                    )
                    assert isinstance(line, str)

            # Test binary mode
            with open(file_name, "rb") as file:
                for idx, line in enumerate(reverse_readline(file)):
                    assert line == contents[len(contents) - idx - 1]

    @pytest.mark.parametrize("file", ["./file", Path("./file")])
    def test_illegal_file_type(self, file):
        with pytest.raises(TypeError, match="expect a file stream, not file name"):
            next(reverse_readline(file))


class TestReverseReadfile:
    NUM_LINES = 3000

    def test_reverse_readfile(self):
        """
        Make sure a file containing line numbers is read in reverse order,
        i.e. the first line that is read corresponds to the last line number.
        """
        fname = os.path.join(TEST_DIR, "3000_lines.txt")
        for idx, line in enumerate(reverse_readfile(fname)):
            assert isinstance(line, str)
            # OS would automatically convert line ending in text mode
            assert line == f"{str(self.NUM_LINES - idx)}{os.linesep}"

    def test_read_gz(self):
        """
        Make sure a file containing line numbers is read in reverse order,
        i.e. the first line that is read corresponds to the last line number.
        """
        fname = os.path.join(TEST_DIR, "3000_lines.txt.gz")
        for idx, line in enumerate(reverse_readfile(fname)):
            assert isinstance(line, str)
            assert line == f"{str(self.NUM_LINES - idx)}\n"

    def test_read_bz2(self):
        """
        Make sure a file containing line numbers is read in reverse order,
        i.e. the first line that is read corresponds to the last line number.
        """
        fname = os.path.join(TEST_DIR, "3000_lines.txt.bz2")
        for idx, line in enumerate(reverse_readfile(fname)):
            assert isinstance(line, str)
            assert line == f"{str(self.NUM_LINES - idx)}\n"

    def test_read_empty_file(self):
        """
        Make sure an empty file does not throw an error when reverse_readline
        is called, which was a problem with an earlier implementation.
        """
        with (
            pytest.warns(match="File is empty, return Unix line ending \n."),
            pytest.warns(match="trying to mmap an empty file"),
        ):
            for _line in reverse_readfile(os.path.join(TEST_DIR, "empty_file.txt")):
                pytest.fail("No error should be thrown.")

    @pytest.mark.parametrize("l_end", ["\n", "\r\n"])
    def test_read_file_with_empty_lines(self, l_end):
        """Empty lines should not be skipped."""
        contents = (f"line1{l_end}", f"{l_end}", f"line3{l_end}")
        filename = "test_empty_line.txt"

        with ScratchDir("."):
            # Test text file
            with open(filename, "w", newline="", encoding="utf-8") as file:
                for line in contents:
                    file.write(line)

            revert_contents = tuple(reverse_readfile(filename))
            assert revert_contents[::-1] == contents

            # Test gzip file
            gzip_filename = f"{filename}.gz"
            with gzip.open(gzip_filename, "w") as file_out:
                for line in contents:
                    file_out.write(line.encode())

            revert_contents_gzip = tuple(reverse_readfile(gzip_filename))
            assert revert_contents_gzip[::-1] == contents

            # Test bzip2 file
            bz2_filename = f"{filename}.bz2"
            with bz2.open(bz2_filename, "w") as file_out:
                for line in contents:
                    file_out.write(line.encode())

            revert_contents_bz2 = tuple(reverse_readfile(bz2_filename))
            assert revert_contents_bz2[::-1] == contents

    @pytest.mark.parametrize("l_end", ["\n", "\r\n"])
    def test_different_line_endings(self, l_end):
        contents = (f"Line1{l_end}", f"Line2{l_end}", f"Line3{l_end}")
        filename = "test_file.txt"

        with ScratchDir("."):
            with open(filename, "w", newline="", encoding="utf-8") as file:
                for line in contents:
                    file.write(line)

            revert_contents = tuple(reverse_readfile(filename))
            assert revert_contents[::-1] == contents


class TestZopen:
    @pytest.mark.parametrize("extension", [".txt", ".bz2", ".gz", ".xz", ".lzma"])
    def test_read_write_files(self, extension):
        """Test read/write in binary/text mode:
        - uncompressed text file: .txt
        - compressed files: bz2/gz/xz/lzma
        """
        filename = f"test_file{extension}"
        content = "This is a test file.\n"

        with ScratchDir("."):
            # Test write and read in text mode
            with zopen(filename, "wt", encoding="utf-8") as f:
                f.write(content)

            with zopen(Path(filename), "rt", encoding="utf-8") as f:
                assert f.read() == content

            # Test write and read in binary mode
            with zopen(Path(filename), "wb") as f:
                f.write(content.encode())

            with zopen(filename, "rb") as f:
                assert f.read() == content.encode()

    def test_lzw_files(self):
        """gzip is not really able to (de)compress LZW files.

        TODO: remove text file real_lzw_file.txt.Z after dropping
        ".Z" extension support
        """
        # Test a fake LZW file (just with .Z extension but DEFLATED algorithm)
        filename = "test.Z"
        content = "This is not a real LZW compressed file.\n"

        with (
            ScratchDir("."),
            pytest.warns(FutureWarning, match="compress LZW-compressed files"),
        ):
            # Test write and read in text mode
            with zopen(filename, "wt", encoding="utf-8") as f:
                f.write(content)

            with zopen(filename, "rt", encoding="utf-8") as f:
                assert f.read() == content

            # Test write and read in binary mode
            with zopen(filename, "wb") as f:
                f.write(content.encode())

            with zopen(filename, "rb") as f:
                assert f.read() == content.encode()

        # Cannot decompress a real LZW file
        with (
            pytest.warns(FutureWarning, match="compress LZW-compressed files"),
            pytest.raises(gzip.BadGzipFile, match="Not a gzipped file"),
            zopen(f"{TEST_DIR}/real_lzw_file.txt.Z", "rt", encoding="utf-8") as f,
        ):
            f.read()

    @pytest.mark.parametrize("extension", [".txt", ".bz2", ".gz", ".xz", ".lzma"])
    def test_warnings(self, extension, monkeypatch):
        filename = f"test_warning{extension}"
        content = "Test warning"

        with ScratchDir("."):
            monkeypatch.setenv("PYTHONWARNDEFAULTENCODING", "1")

            # Default `encoding` warning
            with (
                pytest.warns(EncodingWarning, match="use UTF-8 by default"),
                zopen(filename, "wt") as f,
            ):
                f.write(content)

            # No encoding warning if `PYTHONWARNDEFAULTENCODING` not set
            monkeypatch.delenv("PYTHONWARNDEFAULTENCODING", raising=False)

            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "error",
                    "We strongly encourage explicit `encoding`",
                    EncodingWarning,
                )

                with zopen(filename, "wt") as f:
                    f.write(content)

            # Implicit text/binary `mode` warning
            warnings.filterwarnings(
                "ignore", category=EncodingWarning, message="argument not specified"
            )
            with (
                pytest.warns(
                    FutureWarning, match="discourage using implicit binary/text"
                ),
                zopen(filename, "r") as f,
            ):
                if extension == ".txt":
                    assert f.readline() == content
                else:
                    assert f.readline().decode("utf-8") == content

            # Implicit `mode` warning
            with (
                pytest.warns(FutureWarning, match="discourage using a default `mode`"),
                zopen(filename) as f,
            ):
                if extension == ".txt":
                    assert f.readline() == content
                else:
                    assert f.readline().decode("utf-8") == content


class TestFileLock:
    def setup_method(self):
        self.file_name = "__lock__"
        self.lock = FileLock(self.file_name, timeout=1)
        self.lock.acquire()

    def test_raise(self):
        with pytest.raises(FileLockException):
            new_lock = FileLock(self.file_name, timeout=1)
            new_lock.acquire()

    def teardown_method(self):
        self.lock.release()



from __future__ import annotations

import gc
import pickle
import weakref
from typing import TYPE_CHECKING

import pytest

from monty.design_patterns import cached_class, singleton

if TYPE_CHECKING:
    from typing import Any


class TestSingleton:
    def test_singleton(self):
        @singleton
        class A:
            pass

        a1 = A()
        a2 = A()

        assert id(a1) == id(a2)


@cached_class
class A:
    def __init__(self, val):
        self.val = val

    def __eq__(self, other):
        return self.val == other.val

    def __getinitargs__(self):
        return (self.val,)

    def __getnewargs__(self):
        return (self.val,)


class TestCachedClass:
    def test_cached_class(self):
        a1a = A(1)
        a1b = A(1)
        a2 = A(2)

        assert id(a1a) == id(a1b)
        assert id(a1a) != id(a2)

    def test_pickle(self):
        a = A(2)
        o = pickle.dumps(a)
        assert a == pickle.loads(o)


@cached_class
class ClassForTest:
    def __init__(self, value: Any) -> None:
        self.value = value


def test_caching():
    # Test that instances are cached
    inst1 = ClassForTest(1)
    inst2 = ClassForTest(1)
    assert inst1 is inst2

    inst3 = ClassForTest(2)
    assert inst1 is not inst3


def test_picklability():
    # Test that instances can be pickled and unpickled
    original = ClassForTest(42)
    pickled = pickle.dumps(original)
    unpickled = pickle.loads(pickled)

    # Check that the unpickled instance has the same value
    assert original.value == unpickled.value

    # Check that the unpickled instance is the same as a newly created instance
    new_instance = ClassForTest(42)
    assert unpickled is new_instance


def test_initialization():
    init_count = 0

    @cached_class
    class TestInitClass:
        def __init__(self):
            nonlocal init_count
            init_count += 1

    inst1 = TestInitClass()
    inst2 = TestInitClass()
    assert init_count == 1


def test_class_identity():
    # Ensure the decorated class is still recognized as the original class
    assert isinstance(ClassForTest(1), ClassForTest)
    assert type(ClassForTest(1)) is ClassForTest


def test_multiple_arguments():
    @cached_class
    class MultiArgClass:
        def __init__(self, a, b, c=3):
            self.args = (a, b, c)

    inst1 = MultiArgClass(1, 2)
    inst2 = MultiArgClass(1, 2)
    inst3 = MultiArgClass(1, 2, 4)

    assert inst1 is inst2
    assert inst1 is not inst3
    assert inst1.args == (1, 2, 3)
    assert inst3.args == (1, 2, 4)


def test_different_argument_types():
    @cached_class
    class MultiTypeClass:
        def __init__(self, a, b, c):
            self.a, self.b, self.c = a, b, c

    inst1 = MultiTypeClass(1, "string", (1, 2))
    inst2 = MultiTypeClass(1, "string", (1, 2))
    assert inst1 is inst2

    inst3 = MultiTypeClass(1, "string", [1, 2])  # Unhashable argument
    assert inst1 is not inst3


def test_keyword_arguments():
    @cached_class
    class KeywordClass:
        def __init__(self, a, b, c=3):
            self.a, self.b, self.c = a, b, c

    inst1 = KeywordClass(1, 2)
    inst2 = KeywordClass(a=1, b=2)
    assert inst1 is inst2

    inst3 = KeywordClass(1, 2, 4)
    assert inst1 is not inst3


def test_inheritance_chain():
    @cached_class
    class GrandParent:
        def __init__(self, value):
            self.value = value

    class Parent(GrandParent):
        pass

    @cached_class
    class Child(Parent):
        pass

    gp1 = GrandParent(1)
    gp2 = GrandParent(1)
    assert gp1 is gp2

    p1 = Parent(1)
    p2 = Parent(1)
    assert p1 is p2
    assert p1 is not gp1

    c1 = Child(1)
    c2 = Child(1)
    assert c1 is c2
    assert c1 is not p1


def test_memory_management():
    @cached_class
    class WeakRefClass:
        def __init__(self, value):
            self.value = value

    inst = WeakRefClass(1)
    weak_ref = weakref.ref(inst)

    del inst
    gc.collect()

    assert weak_ref() is None

    # Creating a new instance should work
    new_inst = WeakRefClass(1)
    assert new_inst.value == 1


def test_exception_in_init():
    init_count = 0

    @cached_class
    class ExceptionClass:
        def __init__(self, value):
            nonlocal init_count
            init_count += 1
            if init_count == 1:
                raise ValueError("First init failed")
            self.value = value

    with pytest.raises(ValueError):
        ExceptionClass(1)

    assert init_count == 1

    # Second attempt should work and use cache
    inst1 = ExceptionClass(1)
    inst2 = ExceptionClass(1)
    assert inst1 is inst2
    assert init_count == 2


def test_property_and_method_behavior():
    @cached_class
    class PropertyClass:
        def __init__(self, value):
            self._value = value

        @property
        def value(self):
            return self._value

        def get_value(self):
            return self._value

    inst1 = PropertyClass(1)
    inst2 = PropertyClass(1)

    assert inst1 is inst2
    assert inst1.value == 1
    assert inst1.get_value() == 1


def test_class_method_and_static_method():
    @cached_class
    class MethodClass:
        def __init__(self, value):
            self.value = value

        @classmethod
        def create(cls, value):
            return cls(value)

        @staticmethod
        def static_method(value):
            return value * 2

    inst1 = MethodClass(1)
    inst2 = MethodClass.create(1)
    assert inst1 is inst2

    assert MethodClass.static_method(5) == 10


def test_nested_cached_classes():
    @cached_class
    class OuterClass:
        def __init__(self, value):
            self.value = value

        @cached_class
        class InnerClass:
            def __init__(self, inner_value):
                self.inner_value = inner_value

    outer1 = OuterClass(1)
    outer2 = OuterClass(1)
    assert outer1 is outer2

    inner1 = outer1.InnerClass(2)
    inner2 = outer2.InnerClass(2)
    assert inner1 is inner2


def test_large_number_of_instances():
    @cached_class
    class LargeNumberClass:
        def __init__(self, value):
            self.value = value

    instances = [LargeNumberClass(i) for i in range(1000)]
    for i, inst in enumerate(instances):
        assert inst is LargeNumberClass(i)



from __future__ import annotations

from monty.itertools import chunks, iterator_from_slice


def test_iterator_from_slice():
    assert list(iterator_from_slice(slice(0, 6, 2))) == [0, 2, 4]


def test_chunks():
    assert list(chunks(range(1, 25), 10)) == [
        (1, 2, 3, 4, 5, 6, 7, 8, 9, 10),
        (11, 12, 13, 14, 15, 16, 17, 18, 19, 20),
        (21, 22, 23, 24),
    ]



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import pytest

from monty.os import cd, makedirs_p
from monty.os.path import find_exts, zpath

if TYPE_CHECKING:
    from pathlib import Path

MODULE_DIR = os.path.dirname(__file__)
TEST_DIR = os.path.join(MODULE_DIR, "test_files")


class TestPath:
    def test_zpath_str(self, tmp_path: Path):
        tmp_gz = tmp_path / "tmp.gz"
        tmp_gz.touch()
        ret_path = zpath(str(tmp_gz))
        assert isinstance(ret_path, str)
        assert ret_path == str(tmp_gz)

        tmp_not_bz2 = tmp_path / "tmp_not_bz2"
        tmp_not_bz2.touch()

        ret_path = zpath(f"{tmp_not_bz2}.bz2")
        assert ret_path == str(tmp_not_bz2)
        assert isinstance(ret_path, str)

    def test_zpath_path(self, tmp_path: Path):
        # Test with Path input
        tmp_gz = tmp_path / "tmp.gz"
        tmp_gz.touch()
        ret_path = zpath(tmp_gz)
        assert ret_path == str(tmp_gz)
        assert isinstance(ret_path, str)

    def test_zpath_multiple_extensions(self, tmp_path: Path):
        exts = ["", ".gz", ".GZ", ".bz2", ".BZ2", ".z", ".Z"]
        for ext in exts:
            tmp_file = tmp_path / f"tmp{ext}"
            # create files with all supported compression extensions
            tmp_file.touch()

        # zpath should return the file without compression extension
        ret_path = zpath(tmp_path / "tmp")
        assert ret_path == str(tmp_path / "tmp")
        assert isinstance(ret_path, str)

        (tmp_path / "tmp").unlink()  # Remove the uncompressed file
        ret_path = zpath(tmp_path / "tmp")
        assert ret_path == str(tmp_path / "tmp.gz")  # should find .gz first now

    def test_zpath_nonexistent_file(self, tmp_path: Path):
        # should return path as is for non-existent file
        nonexistent = tmp_path / "nonexistent.txt"
        ret_path = zpath(nonexistent)
        assert ret_path == str(nonexistent)
        ret_path = zpath(f"{nonexistent}.bz2")
        assert ret_path == str(nonexistent)

    def test_find_exts(self):
        assert len(find_exts(MODULE_DIR, "py")) >= 18
        assert len(find_exts(MODULE_DIR, "bz2")) == 2
        n_bz2_excl_tests = len(find_exts(MODULE_DIR, "bz2", exclude_dirs="test_files"))
        assert n_bz2_excl_tests == 0
        n_bz2_w_tests = find_exts(MODULE_DIR, "bz2", include_dirs="test_files")
        assert len(n_bz2_w_tests) == 2


class TestCd:
    def test_cd(self):
        with cd(TEST_DIR):
            assert os.path.exists("empty_file.txt")
        assert not os.path.exists("empty_file.txt")

    def test_cd_exception(self):
        with cd(TEST_DIR):
            assert os.path.exists("empty_file.txt")
        assert not os.path.exists("empty_file.txt")


class TestMakedirs_p:
    def setup_method(self):
        self.test_dir_path = os.path.join(TEST_DIR, "test_dir")

    def test_makedirs_p(self):
        makedirs_p(self.test_dir_path)
        assert os.path.exists(self.test_dir_path)
        makedirs_p(self.test_dir_path)
        with pytest.raises(OSError, match="exists"):
            makedirs_p(os.path.join(TEST_DIR, "3000_lines.txt"))

    def teardown_method(self):
        os.rmdir(self.test_dir_path)



from __future__ import annotations

import logging
from io import StringIO

from monty.logging import logged


@logged()
def add(a, b):
    return a + b


def test_logged():
    s = StringIO()
    logging.basicConfig(level=logging.DEBUG, stream=s)
    add(1, 2)



from __future__ import annotations

from monty.inspect import all_subclasses, caller_name, find_top_pyfile


class LittleCatA:
    pass


class LittleCatB(LittleCatA):
    pass


class LittleCatC:
    pass


class LittleCatD(LittleCatB):
    pass


class TestInspect:
    def test_func(self):
        # Not a real test. Need something better.
        assert find_top_pyfile()
        assert caller_name()

    def test_all_subclasses(self):
        assert all_subclasses(LittleCatA) == [LittleCatB, LittleCatD]



from __future__ import annotations

from monty.bisect import find_ge, find_gt, find_le, find_lt, index


def test_funcs():
    line = [0, 1, 2, 3, 4]
    assert index(line, 1) == 1
    assert find_lt(line, 1) == 0
    assert find_gt(line, 1) == 2
    assert find_le(line, 1) == 1
    assert find_ge(line, 2) == 2
    # assert index([0, 1, 1.5, 2], 1.501, atol=0.1) == 4



from __future__ import annotations

import dataclasses
import datetime
import json
import os
import pathlib
from enum import Enum
from typing import Union

import numpy as np
import pytest

from monty.json import (
    MontyDecoder,
    MontyEncoder,
    MSONable,
    _check_type,
    _load_redirect,
    jsanitize,
    load,
    load2dict,
    partial_monty_encode,
    save,
)

from . import __version__ as TESTS_VERSION

try:
    import pandas as pd
except ImportError:
    pd = None

try:
    import pint
except ImportError:
    pint = None

try:
    import torch
except ImportError:
    torch = None

try:
    import pydantic
except ImportError:
    pydantic = None

try:
    from bson.objectid import ObjectId
except ImportError:
    ObjectId = None

try:
    from bson import json_util
except ImportError:
    json_util = None


TEST_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "test_files")


class GoodMSONClass(MSONable):
    def __init__(self, a, b, c, d=1, *values, **kwargs):
        self.a = a
        self.b = b
        self._c = c
        self._d = d
        self.values = values
        self.kwargs = kwargs

    def __eq__(self, other):
        return (
            self.a == other.a
            and self.b == other.b
            and self._c == other._c
            and self._d == other._d
            and self.kwargs == other.kwargs
            and self.values == other.values
        )


class GoodNOTMSONClass:
    """Literally the same as the GoodMSONClass, except it does not have
    the MSONable inheritance!"""

    def __init__(self, a, b, c, d=1, *values, **kwargs):
        self.a = a
        self.b = b
        self._c = c
        self._d = d
        self.values = values
        self.kwargs = kwargs

    def __eq__(self, other):
        return (
            self.a == other.a
            and self.b == other.b
            and self._c == other._c
            and self._d == other._d
            and self.kwargs == other.kwargs
            and self.values == other.values
        )


class LimitedMSONClass(MSONable):
    """An MSONable class that only accepts a limited number of options"""

    def __init__(self, a):
        self.a = a

    def __eq__(self, other):
        return self.a == other.a


class GoodNestedMSONClass(MSONable):
    def __init__(self, a_list, b_dict, c_list_dict_list, **kwargs):
        assert isinstance(a_list, list)
        assert isinstance(b_dict, dict)
        assert isinstance(c_list_dict_list, list)
        assert isinstance(c_list_dict_list[0], dict)
        first_key = next(iter(c_list_dict_list[0]))
        assert isinstance(c_list_dict_list[0][first_key], list)
        self.a_list = a_list
        self.b_dict = b_dict
        self._c_list_dict_list = c_list_dict_list
        self.kwargs = kwargs


class MethodSerializationClass(MSONable):
    def __init__(self, a):
        self.a = a

    def method(self):
        pass

    @staticmethod
    def staticmethod(self):
        pass

    @classmethod
    def classmethod(cls):
        pass

    def __call__(self, b):
        # override call for instances
        return self.__class__(b)

    class NestedClass:
        def inner_method(self):
            pass


class MethodNonSerializationClass:
    def __init__(self, a):
        self.a = a

    def method(self):
        pass


def my_callable(a, b):
    return a + b


class EnumNoAsDict(Enum):
    name_a = "value_a"
    name_b = "value_b"


class EnumAsDict(Enum):
    name_a = "value_a"
    name_b = "value_b"

    def as_dict(self):
        return {"v": self.value}

    @classmethod
    def from_dict(cls, d):
        return cls(d["v"])


class EnumTest(MSONable, Enum):
    a = 1
    b = 2


class ClassContainingDataFrame(MSONable):
    def __init__(self, df):
        self.df = df


class ClassContainingSeries(MSONable):
    def __init__(self, s):
        self.s = s


class ClassContainingQuantity(MSONable):
    def __init__(self, qty):
        self.qty = qty


class ClassContainingNumpyArray(MSONable):
    def __init__(self, np_a):
        self.np_a = np_a


@dataclasses.dataclass
class Point:
    x: float = 1
    y: float = 2


class Coordinates(MSONable):
    def __init__(self, points):
        self.points = points

    def __str__(self):
        return str(self.points)


@dataclasses.dataclass
class NestedDataClass:
    points: list[Point]


class TestMSONable:
    def setup_method(self):
        self.good_cls = GoodMSONClass

        class BadMSONClass(MSONable):
            def __init__(self, a, b):
                self.a = a
                self.b = b

            def as_dict(self):
                return {"init": {"a": self.a, "b": self.b}}

        self.bad_cls = BadMSONClass

        class BadMSONClass2(MSONable):
            def __init__(self, a, b):
                self.a = a
                self.c = b

        self.bad_cls2 = BadMSONClass2

        class AutoMSON(MSONable):
            def __init__(self, a, b):
                self.a = a
                self.b = b

        self.auto_mson = AutoMSON

        class ClassContainingKWOnlyArgs(MSONable):
            def __init__(self, *, a):
                self.a = a

        self.kw_only_args_cls = ClassContainingKWOnlyArgs

    def test_to_from_dict(self):
        obj = self.good_cls("Hello", "World", "Python")
        d = obj.as_dict()
        assert d is not None
        self.good_cls.from_dict(d)
        jsonstr = obj.to_json()
        d = json.loads(jsonstr)
        assert d["@class"], "GoodMSONClass"
        obj = self.bad_cls("Hello", "World")
        d = obj.as_dict()
        assert d is not None
        with pytest.raises(TypeError):
            self.bad_cls.from_dict(d)
        obj = self.bad_cls2("Hello", "World")
        with pytest.raises(NotImplementedError):
            obj.as_dict()
        obj = self.auto_mson(2, 3)
        d = obj.as_dict()
        self.auto_mson.from_dict(d)

    def test_kw_only_args(self):
        obj = self.kw_only_args_cls(a=1)
        d = obj.as_dict()
        assert d is not None
        assert d["a"] == 1
        self.kw_only_args_cls.from_dict(d)
        jsonstr = obj.to_json()
        d = json.loads(jsonstr)
        assert d["@class"], "ClassContainingKWOnlyArgs"

    def test_unsafe_hash(self):
        GMC = GoodMSONClass
        a_list = [GMC(1, 1.0, "one"), GMC(2, 2.0, "two")]
        b_dict = {"first": GMC(3, 3.0, "three"), "second": GMC(4, 4.0, "four")}
        c_list_dict_list = [
            {
                "list1": [
                    GMC(5, 5.0, "five"),
                    GMC(6, 6.0, "six"),
                    GMC(7, 7.0, "seven"),
                ],
                "list2": [GMC(8, 8.0, "eight")],
            },
            {
                "list3": [
                    GMC(9, 9.0, "nine"),
                    GMC(10, 10.0, "ten"),
                    GMC(11, 11.0, "eleven"),
                    GMC(12, 12.0, "twelve"),
                ],
                "list4": [GMC(13, 13.0, "thirteen"), GMC(14, 14.0, "fourteen")],
                "list5": [GMC(15, 15.0, "fifteen")],
            },
        ]
        obj = GoodNestedMSONClass(
            a_list=a_list, b_dict=b_dict, c_list_dict_list=c_list_dict_list
        )

        assert (
            a_list[0].unsafe_hash().hexdigest()
            == "ea44de0e2ef627be582282c02c48e94de0d58ec6"
        )
        assert (
            obj.unsafe_hash().hexdigest() == "44204c8da394e878f7562c9aa2e37c2177f28b81"
        )

    def test_version(self):
        obj = self.good_cls("Hello", "World", "Python")
        d = obj.as_dict()
        assert d["@version"] == TESTS_VERSION

    def test_nested_to_from_dict(self):
        GMC = GoodMSONClass
        a_list = [GMC(1, 1.0, "one"), GMC(2, 2.0, "two")]
        b_dict = {"first": GMC(3, 3.0, "three"), "second": GMC(4, 4.0, "four")}
        c_list_dict_list = [
            {
                "list1": [
                    GMC(5, 5.0, "five"),
                    GMC(6, 6.0, "six"),
                    GMC(7, 7.0, "seven"),
                ],
                "list2": [GMC(8, 8.0, "eight")],
            },
            {
                "list3": [
                    GMC(9, 9.0, "nine"),
                    GMC(10, 10.0, "ten"),
                    GMC(11, 11.0, "eleven"),
                    GMC(12, 12.0, "twelve"),
                ],
                "list4": [GMC(13, 13.0, "thirteen"), GMC(14, 14.0, "fourteen")],
                "list5": [GMC(15, 15.0, "fifteen")],
            },
        ]
        obj = GoodNestedMSONClass(
            a_list=a_list, b_dict=b_dict, c_list_dict_list=c_list_dict_list
        )

        obj_dict = obj.as_dict()
        obj2 = GoodNestedMSONClass.from_dict(obj_dict)
        assert [obj2.a_list[ii] == aa for ii, aa in enumerate(obj.a_list)]
        assert [obj2.b_dict[kk] == val for kk, val in obj.b_dict.items()]
        assert len(obj.a_list) == len(obj2.a_list)
        assert len(obj.b_dict) == len(obj2.b_dict)
        s = json.dumps(obj_dict)
        obj3 = json.loads(s, cls=MontyDecoder)
        assert [obj2.a_list[ii] == aa for ii, aa in enumerate(obj3.a_list)]
        assert [obj2.b_dict[kk] == val for kk, val in obj3.b_dict.items()]
        assert len(obj3.a_list) == len(obj2.a_list)
        assert len(obj3.b_dict) == len(obj2.b_dict)
        s = json.dumps(obj, cls=MontyEncoder)
        obj4 = json.loads(s, cls=MontyDecoder)
        assert [obj4.a_list[ii] == aa for ii, aa in enumerate(obj.a_list)]
        assert [obj4.b_dict[kk] == val for kk, val in obj.b_dict.items()]
        assert len(obj.a_list) == len(obj4.a_list)
        assert len(obj.b_dict) == len(obj4.b_dict)

    def test_enum_serialization(self):
        e = EnumTest.a
        d = e.as_dict()
        e_new = EnumTest.from_dict(d)
        assert e_new.name == e.name
        assert e_new.value == e.value

        d = {"123": EnumTest.a}
        f = jsanitize(d)
        assert f["123"]["@module"] == "tests.test_json"
        assert f["123"]["@class"] == "EnumTest"
        assert f["123"]["value"] == 1

        f = jsanitize(d, strict=True)
        assert f["123"]["@module"] == "tests.test_json"
        assert f["123"]["@class"] == "EnumTest"
        assert f["123"]["value"] == 1

        f = jsanitize(d, strict=True, enum_values=True)
        assert f["123"] == 1

        f = jsanitize(d, enum_values=True)
        assert f["123"] == 1

    def test_enum_serialization_no_msonable(self):
        d = {"123": EnumNoAsDict.name_a}
        f = jsanitize(d)
        assert f["123"]["@module"] == "tests.test_json"
        assert f["123"]["@class"] == "EnumNoAsDict"
        assert f["123"]["value"] == "value_a"

        f = jsanitize(d, strict=True)
        assert f["123"]["@module"] == "tests.test_json"
        assert f["123"]["@class"] == "EnumNoAsDict"
        assert f["123"]["value"] == "value_a"

        f = jsanitize(d, strict=True, enum_values=True)
        assert f["123"] == "value_a"

        f = jsanitize(d, enum_values=True)
        assert f["123"] == "value_a"

    def test_save_load(self, tmp_path):
        """Tests the save and load serialization methods."""

        test_good_class = GoodMSONClass(
            "Hello",
            "World",
            "Python",
            **{
                "cant_serialize_me": GoodNOTMSONClass(
                    "Hello2", "World2", "Python2", **{"values": []}
                ),
                "cant_serialize_me2": [
                    GoodNOTMSONClass("Hello4", "World4", "Python4", **{"values": []}),
                    GoodNOTMSONClass("Hello4", "World4", "Python4", **{"values": []}),
                ],
                "cant_serialize_me3": [
                    {
                        "tmp": GoodMSONClass(
                            "Hello5", "World5", "Python5", **{"values": []}
                        ),
                        "tmp2": 2,
                        "tmp3": [1, 2, 3],
                    },
                    {
                        "tmp5": GoodNOTMSONClass(
                            "aHello5", "aWorld5", "aPython5", **{"values": []}
                        ),
                        "tmp2": 5,
                        "tmp3": {"test": "test123"},
                    },
                    # Gotta check that if I hide an MSONable class somewhere
                    # it still gets correctly serialized.
                    {"actually_good": GoodMSONClass("1", "2", "3", **{"values": []})},
                ],
                "values": [],
            },
        )

        # This will pass
        test_good_class.as_dict()

        # This will fail
        with pytest.raises(TypeError):
            test_good_class.to_json()

        # This should also pass though
        target = tmp_path / "test.json"
        test_good_class.save(target, json_kwargs={"indent": 4, "sort_keys": True})

        # This will fail
        with pytest.raises(FileExistsError):
            test_good_class.save(target, strict=True)

        # Now check that reloading this, the classes are equal!
        test_good_class2 = GoodMSONClass.load(target)

        # Final check using load
        test_good_class3 = load(target)

        assert test_good_class == test_good_class2
        assert test_good_class == test_good_class3


class TestJson:
    def test_as_from_dict(self):
        obj = GoodMSONClass(1, 2, 3, hello="world")
        s = json.dumps(obj, cls=MontyEncoder)
        obj2 = json.loads(s, cls=MontyDecoder)
        assert obj2.a == 1
        assert obj2.b == 2
        assert obj2._c == 3
        assert obj2._d == 1
        assert obj2.kwargs, {"hello": "world", "values": []}
        obj = GoodMSONClass(obj, 2, 3)
        s = json.dumps(obj, cls=MontyEncoder)
        obj2 = json.loads(s, cls=MontyDecoder)
        assert obj2.a.a == 1
        assert obj2.b == 2
        assert obj2._c == 3
        assert obj2._d == 1
        listobj = [obj, obj2]
        s = json.dumps(listobj, cls=MontyEncoder)
        listobj2 = json.loads(s, cls=MontyDecoder)
        assert listobj2[0].a.a == 1

    @pytest.mark.skipif(torch is None, reason="torch not present")
    def test_torch(self):
        t = torch.tensor([0, 1, 2])
        jsonstr = json.dumps(t, cls=MontyEncoder)
        t2 = json.loads(jsonstr, cls=MontyDecoder)
        assert isinstance(t2, torch.Tensor)
        assert t2.type() == t.type()
        assert np.array_equal(t2, t)
        t = torch.tensor([1 + 1j, 2 + 1j])
        jsonstr = json.dumps(t, cls=MontyEncoder)
        t2 = json.loads(jsonstr, cls=MontyDecoder)
        assert isinstance(t2, torch.Tensor)
        assert t2.type() == t.type()
        assert np.array_equal(t2, t)

    def test_datetime(self):
        dt = datetime.datetime.now()
        jsonstr = json.dumps(dt, cls=MontyEncoder)
        d = json.loads(jsonstr, cls=MontyDecoder)
        assert isinstance(d, datetime.datetime)
        assert dt == d
        # Test a nested datetime.
        a = {"dt": dt, "a": 1}
        jsonstr = json.dumps(a, cls=MontyEncoder)
        d = json.loads(jsonstr, cls=MontyDecoder)
        assert isinstance(d["dt"], datetime.datetime)

        jsanitize(dt, strict=True)

        # test timezone aware datetime API
        created_at = datetime.datetime.now(tz=datetime.timezone.utc)
        data = json.loads(json.dumps(created_at, cls=MontyEncoder))

        created_at_after = MontyDecoder().process_decoded(data)

        assert str(created_at_after).rstrip("0") == str(created_at).rstrip(
            "+00:00"
        ).rstrip("0")

    def test_uuid(self):
        from uuid import UUID, uuid4

        uuid = uuid4()
        jsonstr = json.dumps(uuid, cls=MontyEncoder)
        d = json.loads(jsonstr, cls=MontyDecoder)
        assert isinstance(d, UUID)
        assert uuid == d
        # Test a nested UUID.
        a = {"uuid": uuid, "a": 1}
        jsonstr = json.dumps(a, cls=MontyEncoder)
        d = json.loads(jsonstr, cls=MontyDecoder)
        assert isinstance(d["uuid"], UUID)

    def test_path(self):
        from pathlib import Path

        p = Path("/home/user/")
        jsonstr = json.dumps(p, cls=MontyEncoder)
        d = json.loads(jsonstr, cls=MontyDecoder)
        assert isinstance(d, Path)
        assert d == p

    def test_nan(self):
        x = [float("NaN")]
        djson = json.dumps(x, cls=MontyEncoder)
        d = json.loads(djson)
        assert isinstance(d[0], float)

    def test_numpy(self):
        x = np.array([1, 2, 3], dtype="int64")
        with pytest.raises(TypeError):
            json.dumps(x)
        djson = json.dumps(x, cls=MontyEncoder)
        d = json.loads(djson)
        assert d["@class"] == "array"
        assert d["@module"] == "numpy"
        assert d["data"], [1, 2 == 3]
        assert d["dtype"] == "int64"
        x = json.loads(djson, cls=MontyDecoder)
        assert isinstance(x, np.ndarray)
        x = np.min([1, 2, 3]) > 2
        with pytest.raises(TypeError):
            json.dumps(x)

        x = np.array([1 + 1j, 2 + 1j, 3 + 1j], dtype="complex64")
        with pytest.raises(TypeError):
            json.dumps(x)
        djson = json.dumps(x, cls=MontyEncoder)
        d = json.loads(djson)
        assert d["@class"] == "array"
        assert d["@module"] == "numpy"
        assert d["data"], [[1.0, 2.0, 3.0], [1.0, 1.0 == 1.0]]
        assert d["dtype"] == "complex64"
        x = json.loads(djson, cls=MontyDecoder)
        assert isinstance(x, np.ndarray)
        assert x.dtype == "complex64"

        x = np.array([[1 + 1j, 2 + 1j], [3 + 1j, 4 + 1j]], dtype="complex64")
        with pytest.raises(TypeError):
            json.dumps(x)
        djson = json.dumps(x, cls=MontyEncoder)
        d = json.loads(djson)
        assert d["@class"] == "array"
        assert d["@module"] == "numpy"
        assert d["data"], [[[1.0, 2.0], [3.0, 4.0]], [[1.0, 1.0], [1.0 == 1.0]]]
        assert d["dtype"] == "complex64"
        x = json.loads(djson, cls=MontyDecoder)
        assert isinstance(x, np.ndarray)
        assert x.dtype == "complex64"

        x = {"energies": [np.float64(1234.5)]}
        d = jsanitize(x, strict=True)
        assert isinstance(d["energies"][0], float)

        x = {"energy": np.array(-1.0)}
        d = jsanitize(x, strict=True)
        assert isinstance(d["energy"], float)

        # Test data nested in a class
        x = np.array([[1 + 1j, 2 + 1j], [3 + 1j, 4 + 1j]], dtype="complex64")
        cls = ClassContainingNumpyArray(np_a={"a": [{"b": x}]})

        d = json.loads(json.dumps(cls, cls=MontyEncoder))

        assert d["np_a"]["a"][0]["b"]["@module"] == "numpy"
        assert d["np_a"]["a"][0]["b"]["@class"] == "array"
        assert d["np_a"]["a"][0]["b"]["data"] == [
            [[1.0, 2.0], [3.0, 4.0]],
            [[1.0, 1.0], [1.0, 1.0]],
        ]
        assert d["np_a"]["a"][0]["b"]["dtype"] == "complex64"

        obj = ClassContainingNumpyArray.from_dict(d)
        assert isinstance(obj, ClassContainingNumpyArray)
        assert isinstance(obj.np_a["a"][0]["b"], np.ndarray)
        assert obj.np_a["a"][0]["b"][0][1] == 2 + 1j

    @pytest.mark.skipif(pd is None, reason="pandas not present")
    def test_pandas(self):
        cls = ClassContainingDataFrame(
            df=pd.DataFrame([{"a": 1, "b": 1}, {"a": 1, "b": 2}])
        )

        d = json.loads(MontyEncoder().encode(cls))

        assert d["df"]["@module"] == "pandas"
        assert d["df"]["@class"] == "DataFrame"

        obj = ClassContainingDataFrame.from_dict(d)
        assert isinstance(obj, ClassContainingDataFrame)
        assert isinstance(obj.df, pd.DataFrame)
        assert list(obj.df.a), [1 == 1]

        cls = ClassContainingSeries(s=pd.Series({"a": [1, 2, 3], "b": [4, 5, 6]}))

        d = json.loads(MontyEncoder().encode(cls))

        assert d["s"]["@module"] == "pandas"
        assert d["s"]["@class"] == "Series"

        obj = ClassContainingSeries.from_dict(d)
        assert isinstance(obj, ClassContainingSeries)
        assert isinstance(obj.s, pd.Series)
        assert list(obj.s.a), [1, 2 == 3]

        cls = ClassContainingSeries(
            s={"df": [pd.Series({"a": [1, 2, 3], "b": [4, 5, 6]})]}
        )

        d = json.loads(MontyEncoder().encode(cls))

        assert d["s"]["df"][0]["@module"] == "pandas"
        assert d["s"]["df"][0]["@class"] == "Series"

        obj = ClassContainingSeries.from_dict(d)
        assert isinstance(obj, ClassContainingSeries)
        assert isinstance(obj.s["df"][0], pd.Series)
        assert list(obj.s["df"][0].a), [1, 2 == 3]

    @pytest.mark.skipif(pint is None, reason="pint not present")
    def test_pint_quantity(self):
        ureg = pint.UnitRegistry()
        cls = ClassContainingQuantity(qty=pint.Quantity("9.81 m/s**2"))

        d = json.loads(MontyEncoder().encode(cls))
        assert isinstance(d, dict)

        assert d["qty"]["@module"] == "pint"
        assert d["qty"]["@class"] == "Quantity"
        assert d["qty"].get("@version") is not None

        obj = ClassContainingQuantity.from_dict(d)
        assert isinstance(obj, ClassContainingQuantity)
        assert isinstance(obj.qty, pint.Quantity)
        assert obj.qty.magnitude == 9.81
        assert str(obj.qty.units) == "meter / second ** 2"

    def test_callable(self):
        instance = MethodSerializationClass(a=1)
        for function in [
            # builtins
            str,
            list,
            sum,
            open,
            # functions
            os.path.join,
            my_callable,
            # unbound methods
            MethodSerializationClass.NestedClass.inner_method,
            MethodSerializationClass.staticmethod,
            instance.staticmethod,
            # methods bound to classes
            MethodSerializationClass.classmethod,
            instance.classmethod,
            # classes
            MethodSerializationClass,
            Enum,
        ]:
            with pytest.raises(TypeError):
                json.dumps(function)
            djson = json.dumps(function, cls=MontyEncoder)
            d = json.loads(djson)
            assert "@callable" in d
            assert "@module" in d
            x = json.loads(djson, cls=MontyDecoder)
            assert x == function

        # test method bound to instance
        for function in [instance.method]:
            with pytest.raises(TypeError):
                json.dumps(function)
            djson = json.dumps(function, cls=MontyEncoder)
            d = json.loads(djson)
            assert "@callable" in d
            assert "@module" in d
            x = json.loads(djson, cls=MontyDecoder)

            # can't just check functions are equal as the instance the function is bound
            # to will be different. Instead, we check that the serialized instance
            # is the same, and that the function qualname is the same
            assert x.__qualname__ == function.__qualname__
            assert x.__self__.as_dict() == function.__self__.as_dict()

        # test method bound to object that is not serializable
        for function in [MethodNonSerializationClass(1).method]:
            with pytest.raises(TypeError):
                json.dumps(function, cls=MontyEncoder)

        # test that callable MSONable objects still get serialized as the objects
        # rather than as a callable
        djson = json.dumps(instance, cls=MontyEncoder)
        assert "@class" in djson

    @pytest.mark.skipif(ObjectId is None, reason="bson not present")
    def test_objectid(self):
        oid = ObjectId("562e8301218dcbbc3d7d91ce")
        with pytest.raises(TypeError):
            json.dumps(oid)
        djson = json.dumps(oid, cls=MontyEncoder)
        x = json.loads(djson, cls=MontyDecoder)
        assert isinstance(x, ObjectId)

    def test_jsanitize(self):
        # clean_json should have no effect on None types.
        d = {"hello": 1, "world": None}
        clean = jsanitize(d)
        assert clean["world"] is None
        assert json.loads(json.dumps(d)) == json.loads(json.dumps(clean))

        d = {"hello": GoodMSONClass(1, 2, 3), "test": "hi"}
        with pytest.raises(TypeError):
            json.dumps(d)
        clean = jsanitize(d)
        assert isinstance(clean["hello"], str)
        clean_strict = jsanitize(d, strict=True)
        assert clean_strict["hello"]["a"] == 1
        assert clean_strict["hello"]["b"] == 2
        assert clean_strict["test"] == "hi"
        clean_recursive_msonable = jsanitize(d, recursive_msonable=True)
        assert clean_recursive_msonable["hello"]["a"] == 1
        assert clean_recursive_msonable["hello"]["b"] == 2
        assert clean_recursive_msonable["hello"]["c"] == 3
        assert clean_recursive_msonable["test"] == "hi"

        d = {"hello": [GoodMSONClass(1, 2, 3), "test"], "test": "hi"}
        clean_recursive_msonable = jsanitize(d, recursive_msonable=True)
        assert clean_recursive_msonable["hello"][0]["a"] == 1
        assert clean_recursive_msonable["hello"][0]["b"] == 2
        assert clean_recursive_msonable["hello"][0]["c"] == 3
        assert clean_recursive_msonable["hello"][1] == "test"
        assert clean_recursive_msonable["test"] == "hi"

        d = {"hello": (GoodMSONClass(1, 2, 3), "test"), "test": "hi"}
        clean_recursive_msonable = jsanitize(d, recursive_msonable=True)
        assert clean_recursive_msonable["hello"][0]["a"] == 1
        assert clean_recursive_msonable["hello"][0]["b"] == 2
        assert clean_recursive_msonable["hello"][0]["c"] == 3
        assert clean_recursive_msonable["hello"][1] == "test"
        assert clean_recursive_msonable["test"] == "hi"

        DoubleGoodMSONClass = GoodMSONClass(1, 2, 3)
        DoubleGoodMSONClass.values = [GoodMSONClass(1, 2, 3)]
        clean_recursive_msonable = jsanitize(
            DoubleGoodMSONClass, recursive_msonable=True
        )
        assert clean_recursive_msonable["a"] == 1
        assert clean_recursive_msonable["b"] == 2
        assert clean_recursive_msonable["c"] == 3
        assert clean_recursive_msonable["values"][0]["a"] == 1
        assert clean_recursive_msonable["values"][0]["b"] == 2
        assert clean_recursive_msonable["values"][0]["c"] == 3

        d = {"dt": datetime.datetime.now()}
        clean = jsanitize(d)
        assert isinstance(clean["dt"], str)
        clean = jsanitize(d, allow_bson=True)
        assert isinstance(clean["dt"], datetime.datetime)

        rnd_bin = bytes(np.random.rand(10))
        d = {"a": bytes(rnd_bin)}
        clean = jsanitize(d, allow_bson=True)
        assert clean["a"] == bytes(rnd_bin)
        assert isinstance(clean["a"], bytes)

        p = pathlib.Path("/home/user/")
        clean = jsanitize(p, strict=True)
        assert clean, ["/home/user" in "\\home\\user"]

        # test jsanitizing callables (including classes)
        instance = MethodSerializationClass(a=1)
        for function in [
            # builtins
            str,
            list,
            sum,
            open,
            # functions
            os.path.join,
            my_callable,
            # unbound methods
            MethodSerializationClass.NestedClass.inner_method,
            MethodSerializationClass.staticmethod,
            instance.staticmethod,
            # methods bound to classes
            MethodSerializationClass.classmethod,
            instance.classmethod,
            # classes
            MethodSerializationClass,
            Enum,
        ]:
            d = {"f": function}
            clean = jsanitize(d)
            assert "@module" in clean["f"]
            assert "@callable" in clean["f"]

        # test method bound to instance
        for function in [instance.method]:
            d = {"f": function}
            clean = jsanitize(d)
            assert "@module" in clean["f"]
            assert "@callable" in clean["f"]
            assert clean["f"].get("@bound", None) is not None
            assert "@class" in clean["f"]["@bound"]

        # test method bound to object that is not serializable
        for function in [MethodNonSerializationClass(1).method]:
            d = {"f": function}
            clean = jsanitize(d)
            assert isinstance(clean["f"], str)

            # test that strict checking gives an error
            with pytest.raises(AttributeError):
                jsanitize(d, strict=True)

        # test that callable MSONable objects still get serialized as the objects
        # rather than as a callable
        d = {"c": instance}
        clean = jsanitize(d, strict=True)
        assert "@class" in clean["c"]

    def test_unserializable_composite(self):
        class Unserializable:
            def __init__(self, a):
                self._a = a

            def __str__(self):
                return "Unserializable"

        class Composite(MSONable):
            def __init__(self, name, unserializable, msonable):
                self.name = name
                self.unserializable = unserializable
                self.msonable = msonable

        composite_dictionary = {
            "name": "test",
            "unserializable": Unserializable(1),
            "msonable": GoodMSONClass(1, 2, 3),
        }

        with pytest.raises(AttributeError):
            jsanitize(composite_dictionary, strict=True)

        composite_obj = Composite.from_dict(composite_dictionary)

        with pytest.raises(AttributeError):
            jsanitize(composite_obj, strict=True)

        # Test that skip mode preserves unserializable objects
        skipped_dict = jsanitize(composite_obj, strict="skip", recursive_msonable=True)
        assert skipped_dict["name"] == "test", "String values should remain unchanged"
        assert skipped_dict["unserializable"]._a == 1, (
            "Unserializable object should be preserved in skip mode"
        )
        assert skipped_dict["msonable"]["a"] == 1, (
            "MSONable object should be properly serialized"
        )

        # Test non-strict mode converts unserializable to string
        dict_with_str = jsanitize(composite_obj, strict=False, recursive_msonable=True)
        assert isinstance(dict_with_str["unserializable"], str), (
            "Unserializable object should be converted to string in non-strict mode"
        )

    @pytest.mark.skipif(pd is None, reason="pandas not present")
    def test_jsanitize_pandas(self):
        s = pd.Series({"a": [1, 2, 3], "b": [4, 5, 6]})
        clean = jsanitize(s)
        assert clean == s.to_dict()

    @pytest.mark.skipif(ObjectId is None, reason="bson not present")
    def test_jsanitize_numpy_bson(self):
        d = {
            "a": ["b", np.array([1, 2, 3])],
            "b": ObjectId.from_datetime(datetime.datetime.now()),
        }
        clean = jsanitize(d)
        assert clean["a"], ["b", [1, 2 == 3]]
        assert isinstance(clean["b"], str)

    def test_redirect(self):
        MSONable.REDIRECT["tests.test_json"] = {
            "test_class": {"@class": "GoodMSONClass", "@module": "tests.test_json"},
            "another_test_class": {
                "@class": "AnotherClass",
                "@module": "tests.test_json2",
            },
        }

        d = {
            "@class": "test_class",
            "@module": "tests.test_json",
            "a": 1,
            "b": 1,
            "c": 1,
        }

        obj = json.loads(json.dumps(d), cls=MontyDecoder)
        assert isinstance(obj, GoodMSONClass)

        d2 = {
            "@class": "another_test_class",
            "@module": "tests.test_json",
            "a": 2,
            "b": 2,
            "c": 2,
        }

        with pytest.raises(ImportError, match="No module named 'tests.test_json2'"):
            # This should raise ImportError because it's trying to load
            # AnotherClass from tests.test_json instead of tests.test_json2
            json.loads(json.dumps(d2), cls=MontyDecoder)

    def test_redirect_settings_file(self):
        data = _load_redirect(os.path.join(TEST_DIR, "settings_for_test.yaml"))
        assert data == {
            "old_module": {
                "old_class": {"@class": "new_class", "@module": "new_module"}
            }
        }

    @pytest.mark.skipif(pydantic is None, reason="pydantic not present")
    def test_pydantic_integrations(self):
        from pydantic import BaseModel, ValidationError

        global ModelWithMSONable  # allow model to be deserialized in test
        global LimitedMSONClass

        class ModelWithMSONable(BaseModel):
            a: GoodMSONClass

        test_object = ModelWithMSONable(a=GoodMSONClass(1, 1, 1))
        test_dict_object = ModelWithMSONable(a=test_object.a.as_dict())
        assert test_dict_object.a.a == test_object.a.a
        dict_no_class = test_object.a.as_dict()
        dict_no_class.pop("@class")
        dict_no_class.pop("@module")
        test_dict_object = ModelWithMSONable(a=dict_no_class)
        assert test_dict_object.a.a == test_object.a.a

        assert test_object.model_json_schema() == {
            "title": "ModelWithMSONable",
            "type": "object",
            "properties": {
                "a": {
                    "title": "A",
                    "type": "object",
                    "properties": {
                        "@class": {"enum": ["GoodMSONClass"], "type": "string"},
                        "@module": {"enum": ["tests.test_json"], "type": "string"},
                        "@version": {"type": "string"},
                    },
                    "required": ["@class", "@module"],
                }
            },
            "required": ["a"],
        }

        d = jsanitize(test_object, strict=True, enum_values=True, allow_bson=True)
        assert d == {
            "a": {
                "@module": "tests.test_json",
                "@class": "GoodMSONClass",
                "@version": "0.1",
                "a": 1,
                "b": 1,
                "c": 1,
                "d": 1,
                "values": [],
            },
            "@module": "tests.test_json",
            "@class": "ModelWithMSONable",
            "@version": "0.1",
        }
        obj = MontyDecoder().process_decoded(d)
        assert isinstance(obj, BaseModel)
        assert isinstance(obj.a, GoodMSONClass)
        assert obj.a.b == 1

        # check that a model that is not validated by pydantic still
        # gets completely deserialized.
        global ModelWithDict  # allow model to be deserialized in test

        class ModelWithDict(BaseModel):
            a: dict

        test_object_with_dict = ModelWithDict(a={"x": GoodMSONClass(1, 1, 1)})
        d = jsanitize(
            test_object_with_dict, strict=True, enum_values=True, allow_bson=True
        )
        assert d == {
            "a": {
                "x": {
                    "@module": "tests.test_json",
                    "@class": "GoodMSONClass",
                    "@version": "0.1",
                    "a": 1,
                    "b": 1,
                    "c": 1,
                    "d": 1,
                    "values": [],
                }
            },
            "@module": "tests.test_json",
            "@class": "ModelWithDict",
            "@version": "0.1",
        }
        obj = MontyDecoder().process_decoded(d)
        assert isinstance(obj, BaseModel)
        assert isinstance(obj.a["x"], GoodMSONClass)
        assert obj.a["x"].b == 1

        # check that if an MSONable object raises an exception during
        # the model validation it is properly handled by pydantic
        global ModelWithUnion  # allow model to be deserialized in test
        global ModelWithLimited  # allow model to be deserialized in test

        class ModelWithLimited(BaseModel):
            a: LimitedMSONClass

        class ModelWithUnion(BaseModel):
            a: Union[LimitedMSONClass, dict]

        limited_dict = jsanitize(ModelWithLimited(a=LimitedMSONClass(1)), strict=True)
        assert ModelWithLimited.model_validate(limited_dict)
        limited_dict["a"]["b"] = 2
        with pytest.raises(ValidationError):
            ModelWithLimited.model_validate(limited_dict)

        limited_union_dict = jsanitize(
            ModelWithUnion(a=LimitedMSONClass(1)), strict=True
        )
        validated_model = ModelWithUnion.model_validate(limited_union_dict)
        assert isinstance(validated_model, ModelWithUnion)
        assert isinstance(validated_model.a, LimitedMSONClass)
        limited_union_dict["a"]["b"] = 2
        validated_model = ModelWithUnion.model_validate(limited_union_dict)
        assert isinstance(validated_model, ModelWithUnion)
        assert isinstance(validated_model.a, dict)

    def test_dataclass(self):
        c = Coordinates([Point(1, 2), Point(3, 4)])
        d = c.as_dict()
        c2 = Coordinates.from_dict(d)
        assert d["points"][0]["x"] == 1
        assert d["points"][1]["y"] == 4
        assert isinstance(c2, Coordinates)
        assert isinstance(c2.points[0], Point)

        s = MontyEncoder().encode(Point(1, 2))
        p = MontyDecoder().decode(s)
        assert p.x == 1
        assert p.y == 2

        ndc = NestedDataClass([Point(1, 2), Point(3, 4)])
        str_ = json.dumps(ndc, cls=MontyEncoder)
        ndc2 = json.loads(str_, cls=MontyDecoder)
        assert isinstance(ndc2, NestedDataClass)

    def test_enum(self):
        s = MontyEncoder().encode(EnumNoAsDict.name_a)
        p = MontyDecoder().decode(s)
        assert p.name == "name_a"
        assert p.value == "value_a"

        na1 = EnumAsDict.name_a
        d_ = na1.as_dict()
        assert d_ == {"v": "value_a"}
        na2 = EnumAsDict.from_dict(d_)
        assert na2 == na1

    def test_partial_serializable(self, tmp_path):
        is_m = GoodMSONClass(a=1, b=2, c=3)
        not_m = GoodNOTMSONClass(a="a", b="b", c="c")

        is_m_jsons, is_m_map = partial_monty_encode(is_m)
        is_m_d = json.loads(is_m_jsons)
        assert is_m_d["@class"] == "GoodMSONClass"
        assert is_m_d["a"] == 1
        assert len(is_m_map) == 0

        not_m_jsons, not_m_map = partial_monty_encode(not_m)
        not_m_d = json.loads(not_m_jsons)
        assert not_m_d["@class"] == "GoodNOTMSONClass"
        assert "@object_reference" in not_m_d
        assert len(not_m_map) == 1
        mixed = {"is_m": is_m, "not_m": not_m}
        mixed_jsons, mixed_map = partial_monty_encode(mixed, {"indent": 2})
        mixed_d = json.loads(mixed_jsons)
        assert mixed_d["is_m"]["a"] == 1
        assert mixed_d["is_m"]["@class"] == "GoodMSONClass"
        assert "@object_reference" in mixed_d["not_m"]
        assert mixed_d["not_m"]["@class"] == "GoodNOTMSONClass"

        mixed = {"is_m": is_m, "not_m": not_m}
        save(mixed, tmp_path / "mixed.json")
        loaded = load2dict(tmp_path / "mixed.json")
        assert loaded["is_m"]["a"] == 1
        assert loaded["not_m"].a == "a"

        # Test when you are allowed to overwrite
        with pytest.raises(FileExistsError):
            save(mixed, tmp_path / "mixed.json")

        save(mixed, tmp_path / "mixed.json", strict=False)


class TestCheckType:
    def test_check_subclass(self):
        class A:
            pass

        class B(A):
            pass

        a, b = A(), B()

        class_name_A = f"{type(a).__module__}.{type(a).__qualname__}"
        class_name_B = f"{type(b).__module__}.{type(b).__qualname__}"

        # a is an instance of A, but not B
        assert _check_type(a, class_name_A)
        assert isinstance(a, A)
        assert not _check_type(a, class_name_B)
        assert not isinstance(a, B)

        # b is an instance of both B and A
        assert _check_type(b, class_name_B)
        assert isinstance(b, B)
        assert _check_type(b, class_name_A)
        assert isinstance(b, A)

    def test_check_class(self):
        """This should not work for classes."""

        class A:
            pass

        class B(A):
            pass

        class_name_A = f"{A.__module__}.{A.__qualname__}"
        class_name_B = f"{B.__module__}.{B.__qualname__}"

        # Test class behavior (should return False, like isinstance does)
        assert not _check_type(A, class_name_A)
        assert not _check_type(B, class_name_B)
        assert not _check_type(B, class_name_A)

    def test_callable(self):
        # Test function
        def my_function():
            pass

        callable_class_name = (
            f"{type(my_function).__module__}.{type(my_function).__qualname__}"
        )

        assert _check_type(my_function, callable_class_name), callable_class_name
        assert isinstance(my_function, type(my_function))

        # Test callable class
        class MyCallableClass:
            def __call__(self):
                pass

        callable_instance = MyCallableClass()
        assert callable(callable_instance)

        callable_class_instance_name = f"{type(callable_instance).__module__}.{type(callable_instance).__qualname__}"

        assert _check_type(callable_instance, callable_class_instance_name), (
            callable_class_instance_name
        )
        assert isinstance(callable_instance, MyCallableClass)

    def test_numpy(self):
        # Test NumPy array
        arr = np.array([1, 2, 3])

        assert _check_type(arr, "numpy.ndarray")
        assert isinstance(arr, np.ndarray)

        # Test NumPy generic
        scalar = np.float64(3.14)

        assert _check_type(scalar, "numpy.generic")
        assert isinstance(scalar, np.generic)

    @pytest.mark.skipif(pd is None, reason="pandas is not installed")
    def test_pandas(self):
        # Test pandas DataFrame
        df = pd.DataFrame({"a": [1, 2, 3]})

        assert _check_type(df, "pandas.core.frame.DataFrame")
        assert isinstance(df, pd.DataFrame)

        assert _check_type(df, "pandas.core.base.PandasObject")
        assert isinstance(df, pd.core.base.PandasObject)

        # Test pandas Series
        series = pd.Series([1, 2, 3])

        assert _check_type(series, "pandas.core.series.Series")
        assert isinstance(series, pd.Series)

        assert _check_type(series, "pandas.core.base.PandasObject")
        assert isinstance(series, pd.core.base.PandasObject)

    @pytest.mark.skipif(torch is None, reason="torch is not installed")
    def test_torch(self):
        tensor = torch.tensor([1, 2, 3])

        assert _check_type(tensor, "torch.Tensor")
        assert isinstance(tensor, torch.Tensor)

    @pytest.mark.skipif(pydantic is None, reason="pydantic is not installed")
    def test_pydantic(self):
        class MyModel(pydantic.BaseModel):
            name: str

        model_instance = MyModel(name="Alice")

        assert _check_type(model_instance, "pydantic.main.BaseModel")
        assert isinstance(model_instance, pydantic.BaseModel)

    @pytest.mark.skipif(pint is None, reason="pint is not installed")
    def test_pint(self):
        ureg = pint.UnitRegistry()
        qty = 3 * ureg.meter

        assert _check_type(qty, "pint.registry.Quantity")
        assert isinstance(qty, pint.Quantity)

    @pytest.mark.skipif(ObjectId is None, reason="bson not present")
    @pytest.mark.skipif(json_util is None, reason="pymongo not present")
    def test_extended_json(self):
        ext_json_dict = {
            "datetime": datetime.datetime.now(datetime.timezone.utc),
            "NaN": float("NaN"),
            "infinity": float("inf"),
            "-infinity": -float("inf"),
        }
        ext_json_str = json_util.dumps(ext_json_dict)

        not_serialized = json.loads(ext_json_str)
        assert all(isinstance(v, dict) for v in not_serialized.values())

        reserialized = MontyDecoder().decode(ext_json_str)
        for k, v in ext_json_dict.items():
            if k == "datetime":
                # BSON's json_util only saves datetimes up to microseconds
                assert reserialized[k].timestamp() == pytest.approx(
                    v.timestamp(), abs=1e-3
                )
            elif k == "NaN":
                assert np.isnan(reserialized[k])
            else:
                assert v == reserialized[k]
            assert not isinstance(reserialized[k], dict)



from __future__ import annotations

import os

from monty.re import regrep

TEST_DIR = os.path.join(os.path.dirname(__file__), "test_files")


def test_regrep():
    """
    We are making sure a file containing line numbers is read in reverse
    order, i.e. the first line that is read corresponds to the last line.
    number
    """
    fname = os.path.join(TEST_DIR, "3000_lines.txt")
    matches = regrep(fname, {"1": r"1(\d+)", "3": r"3(\d+)"}, postprocess=int)
    assert len(matches["1"]) == 1380
    assert len(matches["3"]) == 571
    assert matches["1"][0][0][0] == 0

    matches = regrep(
        fname,
        {"1": r"1(\d+)", "3": r"3(\d+)"},
        reverse=True,
        terminate_on_match=True,
        postprocess=int,
    )
    assert len(matches["1"]) == 1
    assert len(matches["3"]) == 11



from __future__ import annotations

import unittest
import warnings
from dataclasses import dataclass

import pytest

from monty.dev import deprecated, install_excepthook, requires

try:
    from IPython.core import ultratb
except ImportError:
    ultratb = None

# Set all warnings to always be triggered.
warnings.simplefilter("always")


class TestDeprecated:
    def test_basic_usage(self):
        def func_replace():
            pass

        @deprecated(func_replace, "Use func_replace instead", deadline=(2025, 1, 1))
        def func_old():
            """This is the old function."""
            pass

        with warnings.catch_warnings(record=True) as w:
            # Trigger a warning.
            func_old()

            assert issubclass(w[0].category, FutureWarning)
            assert "Use func_replace instead" in str(w[0].message)
            assert "will be removed on 2025-01-01" in str(w[0].message)

        # Check metadata preservation
        assert func_old.__name__ == "func_old"
        assert func_old.__doc__ == "This is the old function."

    def test_str_replacement(self):
        @deprecated("func_replace")
        def func_old():
            pass

        with warnings.catch_warnings(record=True) as w:
            # Trigger a warning.
            func_old()
            # Verify Warning and message
            assert issubclass(w[0].category, FutureWarning)
            assert "use func_replace instead" in str(w[0].message)

    def test_property(self):
        class TestClass:
            """A dummy class for tests."""

            @property
            def property_a(self):
                pass

            @property
            @deprecated(property_a)
            def property_b(self):
                return "b"

            @deprecated(property_a)
            def func_a(self):
                return "a"

        with warnings.catch_warnings(record=True) as w:
            # Trigger a warning.
            assert TestClass().property_b == "b"
            # Verify warning type
            assert issubclass(w[-1].category, FutureWarning)

        with warnings.catch_warnings(record=True) as w:
            # Trigger a warning.
            assert TestClass().func_a() == "a"
            # Verify some things
            assert issubclass(w[-1].category, FutureWarning)

    def test_classmethod(self):
        class TestClass:
            """A dummy class for tests."""

            @classmethod
            def classmethod_a(cls):
                pass

            @classmethod
            @deprecated(classmethod_a)
            def classmethod_b(cls):
                return "b"

        with warnings.catch_warnings(record=True) as w:
            # Trigger a warning.
            assert TestClass().classmethod_b() == "b"
            # Verify some things
            assert issubclass(w[-1].category, FutureWarning)

        class TestClass_deprecationwarning:
            """A dummy class for tests."""

            @classmethod
            def classmethod_a(cls):
                pass

            @classmethod
            @deprecated(classmethod_a, category=DeprecationWarning)
            def classmethod_b(cls):
                return "b"

        with pytest.warns(DeprecationWarning):
            assert TestClass_deprecationwarning().classmethod_b() == "b"

    def test_class(self):
        class TestClassNew:
            """A dummy class for tests."""

            def method_a(self):
                pass

        @deprecated(replacement=TestClassNew)
        class TestClassOld:
            """A dummy old class for tests."""

            class_attrib_old = "OLD_ATTRIB"

            def method_b(self):
                """This is method_b."""
                pass

        with pytest.warns(FutureWarning, match="TestClassOld is deprecated"):
            old_class = TestClassOld()

        # Check metadata preservation
        assert old_class.__doc__ == "A dummy old class for tests."
        assert old_class.class_attrib_old == "OLD_ATTRIB"
        assert old_class.__module__ == __name__

        assert old_class.method_b.__doc__ == "This is method_b."

    def test_dataclass(self):
        @dataclass
        class TestClassNew:
            """A dummy class for tests."""

            def __post_init__(self):
                pass

            def method_a(self):
                pass

        @deprecated(replacement=TestClassNew)
        @dataclass
        class TestClassOld:
            """A dummy old class for tests."""

            class_attrib_old = "OLD_ATTRIB"

            def __post_init__(self):
                pass

            def method_b(self):
                """This is method_b."""
                pass

        with pytest.warns(FutureWarning, match="TestClassOld is deprecated"):
            old_class = TestClassOld()

        # Check metadata preservation
        assert old_class.__doc__ == "A dummy old class for tests."
        assert old_class.class_attrib_old == "OLD_ATTRIB"


def test_requires():
    try:
        import fictitious_mod
    except ImportError:
        fictitious_mod = None

    err_msg = "fictitious_mod is not present."

    @requires(fictitious_mod is not None, err_msg)
    def use_fictitious_mod():
        pass

    with pytest.raises(RuntimeError, match=err_msg):
        use_fictitious_mod()

    @requires(unittest is not None, "unittest is not present.")
    def use_unittest():
        return "success"

    assert use_unittest() == "success"

    # test with custom error class
    @requires(False, "expect ImportError", err_cls=ImportError)
    def use_import_error():
        return "success"

    with pytest.raises(ImportError, match="expect ImportError"):
        use_import_error()


@pytest.mark.skipif(ultratb is None, reason="ipython is not installed")
def test_install_except_hook():
    install_excepthook()
